import platform
from pathlib import Path
from functools import wraps

import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger

import pcleaner.config as cfg
import pcleaner.gui.state_saver as ss
from pcleaner.gui.ui_generated_files.ui_PostActionConfiguration import Ui_PostActionConfiguration

SHUTDOWN_COMMAND_NAME = "SHUTDOWN"


def defer_button_update(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        self.disable_button_check = True
        result = method(self, *args, **kwargs)
        self.disable_button_check = False
        self.update_button_enabled()
        return result

    return wrapper


class PostActionConfiguration(Qw.QDialog, Ui_PostActionConfiguration):
    """
    The post action configuration dialog.
    """

    config: cfg.Config
    disable_button_check: bool

    state_saver: ss.StateSaver  # The state saver for the window.

    def __init__(
        self,
        parent=None,
        config: cfg.Config = None,
    ) -> None:
        """
        Init the widget.

        :param parent: The parent widget.
        """
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.config = config
        self.disable_button_check = False

        self.checkBox_remember_action.setChecked(config.pa_remember_action)
        self.checkBox_cancel_on_error.setChecked(config.pa_cancel_on_error)
        self.spinBox_wait_time.setValue(config.pa_wait_time)
        self.lineEdit_shutdown.setText(config.pa_shutdown_command or "")
        self.lineEdit_shutdown.setPlaceholderText(shut_down_command())

        for name, command in config.pa_custom_commands.items():
            self.tableWidget_commands.appendRow(name, command)

        self.pushButton_new.clicked.connect(
            lambda: self.tableWidget_commands.appendRow("", "", select_new=True)
        )
        self.pushButton_delete.clicked.connect(self.delete_command)
        self.pushButton_row_up.clicked.connect(self.move_row_up)
        self.pushButton_row_down.clicked.connect(self.move_row_down)
        self.tableWidget_commands.currentRowChanged.connect(self.update_button_enabled)
        self.tableWidget_commands.cellChanged.connect(self.update_button_enabled)

        self.label_warning.hide()
        self.label_warning_icon.hide()
        self.label_warning_icon.setPixmap(Qg.QIcon.fromTheme("data-warning").pixmap(16, 16))

        self.update_button_enabled()

        self.state_saver = ss.StateSaver("post_action_config")
        self.init_state_saver()
        self.state_saver.restore()

    def init_state_saver(self) -> None:
        """
        Load the state from the state saver.
        """
        self.state_saver.register(
            self,
            self.tableWidget_commands,
        )

    def closeEvent(self, event: Qg.QCloseEvent) -> None:
        self.state_saver.save()
        event.accept()

    def delete_command(self) -> None:
        """
        Delete the selected command.
        """
        row = self.tableWidget_commands.currentRow()
        if row == -1:
            return
        self.tableWidget_commands.removeRow(row)
        self.update_button_enabled()

    @defer_button_update
    def move_row_up(self) -> None:
        row = self.tableWidget_commands.currentRow()
        if row == -1:
            return
        self.tableWidget_commands.moveRowUp(row)

    @defer_button_update
    def move_row_down(self) -> None:
        row = self.tableWidget_commands.currentRow()
        if row == -1:
            return
        self.tableWidget_commands.moveRowDown(row)

    def update_button_enabled(self) -> None:
        """
        Update the enabled state of the buttons.
        """
        if self.disable_button_check:
            return
        row = self.tableWidget_commands.currentRow()
        self.pushButton_delete.setEnabled(row != -1)
        self.pushButton_row_up.setEnabled(row > 0)
        self.pushButton_row_down.setEnabled(
            row != -1 and row < self.tableWidget_commands.rowCount() - 1
        )
        self.buttonBox.button(Qw.QDialogButtonBox.Ok).setEnabled(self.commands_valid())

    def commands_valid(self) -> bool:
        """
        Check if the commands are valid. All we care about is that the names are not empty
        and unique.
        """
        names = set()
        for row in range(self.tableWidget_commands.rowCount()):
            name = self.tableWidget_commands.item(row, 0).text()
            if not name:
                self.label_warning.setText(self.tr("Action names cannot be empty."))
                self.label_warning.show()
                self.label_warning_icon.show()
                return False
            if name in names:
                self.label_warning.setText(self.tr("Action names must be unique."))
                self.label_warning.show()
                self.label_warning_icon.show()
                return False
            if name == SHUTDOWN_COMMAND_NAME:
                self.label_warning.setText(
                    self.tr('The name "{shutdown_command}" is reserved.').format(
                        shutdown_command=SHUTDOWN_COMMAND_NAME
                    )
                )
                self.label_warning.show()
                self.label_warning_icon.show()
                return False
            # Make sure this isn't the flatpak version:
            if Path("/.flatpak-info").exists():
                self.label_warning.setText(
                    self.tr("The Flatpak sandbox will likely prevent actions from working.")
                )
                self.label_warning.show()
                self.label_warning_icon.show()
                return False

            names.add(name)
        self.label_warning.setText("")
        self.label_warning.hide()
        self.label_warning_icon.hide()
        return True

    def accept(self):
        """
        Accept the dialog.
        """
        if not self.commands_valid():  # Just to be sure.
            return
        self.config.pa_remember_action = self.checkBox_remember_action.isChecked()
        self.config.pa_cancel_on_error = self.checkBox_cancel_on_error.isChecked()
        self.config.pa_wait_time = self.spinBox_wait_time.value()
        self.config.pa_shutdown_command = self.lineEdit_shutdown.text()

        self.config.pa_custom_commands = {
            self.tableWidget_commands.item(row, 0)
            .text(): self.tableWidget_commands.item(row, 1)
            .text()
            for row in range(self.tableWidget_commands.rowCount())
        }

        self.config.save()

        super().accept()


def shut_down_command() -> str:
    if platform.system() == "Windows":
        return "shutdown /s /t 0"
    else:  # Works on all Unix-like systems.
        return "shutdown -h now"



import torch
from PySide6.QtCore import Signal, QObject, QTimer

import pcleaner.helpers as hp


class MemoryWatcher(QObject):
    # Monitor system memory usage to warn of impending OOM errors.
    oom_warning = Signal(str)
    oom_relaxed = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.check_memory)

    def start(self):
        # Start the timer to check memory every second (1000 ms)
        self.timer.start(2000)

    def stop(self):
        # Stop the timer
        self.timer.stop()

    def check_memory(self):
        # Monitor system memory (RAM)
        oom: bool = False

        # Check conditions for RAM and swap.
        if hp.sys_swap_memory_total() == 0:  # No swap available.
            if (mem := hp.sys_virtual_memory_used_percent()) >= 80:
                self.oom_warning.emit(
                    self.tr("RAM usage has reached {mem}%").format(mem=round(mem))
                )
                oom = True
        else:  # Swap is available
            if (mem := hp.sys_virtual_memory_used_percent()) >= 90:
                # self.oom_warning.emit("RAM usage has reached 100%.")
                self.oom_warning.emit(
                    self.tr("RAM usage has reached {mem}%").format(mem=round(mem))
                )
                oom = True

        # Monitor VRAM if torch is available and GPU is detected.
        if torch.cuda.is_available() and torch.cuda.device_count() > 0:
            vram_total = torch.cuda.get_device_properties(0).total_memory
            vram_used = torch.cuda.memory_allocated(0)
            vram_percent = (vram_used / vram_total) * 100

            if vram_percent >= 80:
                self.oom_warning.emit(
                    "VRAM usage has reached {vmem}%".format(vmem=round(vram_percent))
                )
                oom = True

        if not oom:
            self.oom_relaxed.emit()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'FileManagerIntegration.ui'
##
## Created by: Qt User Interface Compiler version 6.8.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCommandLinkButton, QDialog,
    QDialogButtonBox, QLabel, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)

class Ui_FileManagerExtension(object):
    def setupUi(self, FileManagerExtension):
        if not FileManagerExtension.objectName():
            FileManagerExtension.setObjectName(u"FileManagerExtension")
        FileManagerExtension.resize(650, 420)
        self.verticalLayout = QVBoxLayout(FileManagerExtension)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(12, 12, 12, -1)
        self.label = QLabel(FileManagerExtension)
        self.label.setObjectName(u"label")
        self.label.setTextFormat(Qt.MarkdownText)
        self.label.setWordWrap(True)

        self.verticalLayout.addWidget(self.label)

        self.verticalSpacer = QSpacerItem(20, 12, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.commandLinkButton_install = QCommandLinkButton(FileManagerExtension)
        self.commandLinkButton_install.setObjectName(u"commandLinkButton_install")
        self.commandLinkButton_install.setText(u"<add shell extension>")
        self.commandLinkButton_install.setDescription(u"<adding to what program>")

        self.verticalLayout.addWidget(self.commandLinkButton_install)

        self.verticalSpacer_3 = QSpacerItem(20, 12, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

        self.verticalLayout.addItem(self.verticalSpacer_3)

        self.commandLinkButton_uninstall = QCommandLinkButton(FileManagerExtension)
        self.commandLinkButton_uninstall.setObjectName(u"commandLinkButton_uninstall")
        self.commandLinkButton_uninstall.setText(u"<remove shell extension>")
        icon = QIcon()
        iconThemeName = u"edit-delete"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.commandLinkButton_uninstall.setIcon(icon)
        self.commandLinkButton_uninstall.setDescription(u"<removing from what program>")

        self.verticalLayout.addWidget(self.commandLinkButton_uninstall)

        self.verticalSpacer_2 = QSpacerItem(20, 12, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.buttonBox = QDialogButtonBox(FileManagerExtension)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setStandardButtons(QDialogButtonBox.Close)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(FileManagerExtension)
        self.buttonBox.rejected.connect(FileManagerExtension.reject)

        QMetaObject.connectSlotsByName(FileManagerExtension)
    # setupUi

    def retranslateUi(self, FileManagerExtension):
        FileManagerExtension.setWindowTitle(QCoreApplication.translate("FileManagerExtension", u"File Manager Extension", None))
        self.label.setText(QCoreApplication.translate("FileManagerExtension", u"### Add or Remove Integration with your File Browser\n"
"\n"
"This feature adds a menu option to your file browser's context menu (which appears when you right-click) for folders or image files.\n"
"\n"
"It allows you to use Panel Cleaner directly from the file browser, eliminating the need to open the application separately.\n"
"\n"
"By default, this uses the built-in profile. However, you can change this default setting by choosing **Profile** and then **Set Default** from the application's menu bar.", None))
    # retranslateUi




from enum import Enum, IntEnum, auto
from pathlib import Path
from typing import Protocol, Any
from uuid import uuid4

from attrs import frozen


class Output(IntEnum):
    """
    These enums represent all the output images displayed in the gui.
    These are an IntEnum so that they can be compared by order.
    This way, we can determine the last step in a set of outputs.
    """

    input = auto()
    ai_mask = auto()
    raw_boxes = auto()
    raw_json = auto()

    initial_boxes = auto()
    final_boxes = auto()
    ocr = auto()
    clean_json = auto()

    box_mask = auto()
    cut_mask = auto()
    mask_layers = auto()
    fitment_quality = auto()
    mask_overlay = auto()
    final_mask = auto()
    isolated_text = auto()
    masked_output = auto()
    mask_data_json = auto()

    denoise_mask = auto()
    denoised_output = auto()

    inpainted_mask = auto()
    inpainted_output = auto()

    write_output = auto()  # This is only used for the progress bar.


class Step(IntEnum):
    """
    These enums represent all the steps in the image processing pipeline.
    """

    text_detection = 1
    preprocessor = auto()
    masker = auto()
    denoiser = auto()
    inpainter = auto()
    output = auto()


class ProgressType(Enum):
    """
    The type of progress reported:
    - Begin: The step started, so the progress bar should be reset.
    - Incremental: One (or more) tasks were completed.
    - Absolute: The progress bar should be set to a specific value.
    - End: The step finished, so the progress bar should be set to 100%.
    """

    start = auto()
    begin_step = auto()
    incremental = auto()
    absolute = auto()
    textDetection_done = auto()
    analyticsOCR = auto()
    analyticsMasker = auto()
    analyticsDenoiser = auto()
    analyticsInpainter = auto()
    outputOCR = auto()
    end = auto()
    aborted = auto()


@frozen
class ProgressData:
    """
    A callback to report progress to the gui.

    Consists of:
    - The total number of images to process.
    - The target output to reach.
    - The current step.
    - The progress type.
    - A value for the case of absolute progress, or how much to increment by for incremental progress, or whatever analytics.
    """

    total_images: int
    target_outputs: list[Output]
    current_step: Step
    progress_type: ProgressType
    value: Any = 1  # Default value for incremental progress. Otherwise usually an analytics struct.


class ProgressSignal(Protocol):
    def emit(self, data: ProgressData) -> None:
        pass


output_to_step: dict[Output, Step] = {
    Output.input: Step.text_detection,
    Output.ai_mask: Step.text_detection,
    Output.raw_boxes: Step.text_detection,
    Output.raw_json: Step.text_detection,
    Output.initial_boxes: Step.preprocessor,
    Output.final_boxes: Step.preprocessor,
    Output.ocr: Step.preprocessor,
    Output.clean_json: Step.preprocessor,
    Output.box_mask: Step.masker,
    Output.cut_mask: Step.masker,
    Output.mask_layers: Step.masker,
    Output.fitment_quality: Step.masker,
    Output.mask_overlay: Step.masker,
    Output.final_mask: Step.masker,
    Output.isolated_text: Step.masker,
    Output.masked_output: Step.masker,
    Output.mask_data_json: Step.masker,
    Output.denoise_mask: Step.denoiser,
    Output.denoised_output: Step.denoiser,
    Output.inpainted_mask: Step.inpainter,
    Output.inpainted_output: Step.inpainter,
    Output.write_output: Step.output,
}

# The final output representing each step.
# If this output is intact, consider the step complete.
step_to_output: dict[Step, tuple[Output, ...]] = {
    Step.text_detection: (Output.input, Output.ai_mask, Output.raw_boxes, Output.raw_json),
    Step.preprocessor: (Output.initial_boxes, Output.final_boxes, Output.ocr, Output.clean_json),
    Step.masker: (
        Output.box_mask,
        Output.cut_mask,
        Output.mask_layers,
        Output.fitment_quality,
        Output.mask_overlay,
        Output.isolated_text,
        Output.masked_output,
        Output.final_mask,
        Output.mask_data_json,
    ),
    Step.denoiser: (Output.denoise_mask, Output.denoised_output),
    Step.inpainter: (Output.inpainted_mask, Output.inpainted_output),
    Step.output: (Output.write_output,),
}

# When checking which outputs are affected by a profile change, we can check
# from the last step to the first step, and stop when we find an output that is
# unaffected by the profile change.
# But these outputs are affected by changes that the ones after them are not,
# so we can't use that optimization.
# In the case of mask_overlay, it's affected by the debug mask color, but nothing else is.
OUTPUTS_WITH_INDEPENDENT_PROFILE_SENSITIVITY = (Output.mask_overlay,)


class ImageAnalyticCategory(Enum):
    """
    The category of analytics.
    """

    ocr_removed = auto()
    mask_failed = auto()
    mask_perfect = auto()
    denoised = auto()
    inpainted = auto()


class ImageAnalytics:
    """
    Bundle all the necessary info to display the minimal analytics for an image.
    These numbers are shown below their corresponding icon in the file table.
    total means the number of boxes or masks from which the failed etc. numbers are derived.
    This total can shrink as the image is processed, since if a box is thrown out by ocr, there
    won't be a corresponding mask for it etc., hence the seemingly redundant multiple totals.

    The analytics per category are stored as simple strings in the form "value/total".

    If the analytic value is blank, it should be hidden in the gui, since it's not relevant.
    These analytics are supposed to give a quick overview of how bad or good the image is, or where
    problems might be.
    """

    _data: dict[ImageAnalyticCategory, str]

    def __init__(self) -> None:
        self._data = {category: "" for category in ImageAnalyticCategory}

    def get_category(self, category: ImageAnalyticCategory) -> str:
        return self._data[category]

    def set_category(self, category: ImageAnalyticCategory, value: int, total: int | None) -> None:
        """
        Set the analytic value for the category as a string internally.
        If either value is 0, the analytic is reset.

        :param category: The category to set the value for.
        :param value: The value to set.
        :param total: The total to set.
        """
        if value == 0 or total == 0:
            self._data[category] = ""
        elif total is None:
            self._data[category] = str(value)
        else:
            self._data[category] = f"{value}/{total}"


def get_output_representing_step(step: Step) -> Output:
    """
    Get the output representing the step, which is the last one in the pipeline.
    This is the json output for the first three steps, which forms the foundation for the
    next step to continue from.

    :param step: The step to get the output for.
    :return: The output representing the step.
    """
    return step_to_output[step][-1]


class OutputPathGenerator:
    """
    Handle creating various output paths for all the different output options.
    """

    original_path: Path
    output_dir: Path
    uuid: str

    # A suffix-free str that acts as a prefix for all cached file paths.
    _output_base_path: str

    def __init__(
        self,
        original_path: Path,
        output_dir: Path,
        uuid_source: Path | str | None = None,
        export_mode: bool = False,
    ) -> None:
        """
        Provide the original image path and cache dir to generate subsequent paths.
        Optionally, provide a path that contains the uuid to use the same one, otherwise
        a new uuid for clobber protection is generated.

        :param original_path: Image path of the input image.
        :param output_dir: The cache directory that intermediate images end up in.
        :param uuid_source: [Optional] a path with the uuid or the uuid as a string.
        :param export_mode: [Optional] when True, don't attempt to add a uuid.
        """
        self.original_path = original_path.resolve()
        self.output_dir = output_dir.resolve()
        if uuid_source is None:
            uuid = uuid4()
        else:
            if isinstance(uuid_source, str):
                uuid = uuid_source
            elif isinstance(uuid_source, Path):
                # uuids are used as clobber protection, like: "{UUID}_filename_suffix.extension"
                # ex. d91d86d1-b8d2-400b-98b2-2d0337973631_0023_myimage_base.png
                uuid = uuid_source.stem.split("_")[0]
            else:
                raise TypeError("uuid_source expected to be Path, str, or None")

        if not export_mode:
            self._output_base_path = str(output_dir / f"{uuid}_{original_path.stem}")
        else:
            self._output_base_path = str(output_dir / original_path.stem)

        self.uuid = uuid

    def for_output(self, output: Output) -> Path:
        match output:
            case Output.input:
                return self.base_png
            case Output.ai_mask:
                return self.raw_mask
            case Output.raw_boxes:
                return self.raw_boxes
            case Output.raw_json:
                return self.raw_json
            case Output.initial_boxes:
                return self.boxes
            case Output.final_boxes:
                return self.final_boxes
            case Output.clean_json:
                return self.clean_json
            case Output.box_mask:
                return self.box_mask
            case Output.cut_mask:
                return self.cut_mask
            case Output.mask_layers:
                return self.mask_fitments
            case Output.final_mask:
                return self.combined_mask
            case Output.mask_overlay:
                return self.mask_overlay
            case Output.fitment_quality:
                return self.std_devs
            case Output.isolated_text:
                return self.text
            case Output.masked_output:
                return self.clean
            case Output.mask_data_json:
                return self.mask_data_json
            case Output.denoise_mask:
                return self.noise_mask
            case Output.denoised_output:
                return self.clean_denoised
            case Output.inpainted_mask:
                return self.inpainting
            case Output.inpainted_output:
                return self.clean_inpaint
            case _:
                raise ValueError(f"No path assigned for output {output}")

    def _attach(self, suffix: str) -> Path:
        return Path(self._output_base_path + suffix)

    @property
    def base_png(self) -> Path:
        return self._attach("_base.png")

    @property
    def raw_mask(self) -> Path:
        return self._attach("_raw_mask.png")

    @property
    def mask(self) -> Path:
        return self._attach("_mask.png")

    @property
    def raw_boxes(self) -> Path:
        return self._attach("_raw_boxes.png")

    @property
    def raw_json(self) -> Path:
        return self._attach("#raw.json")

    @property
    def boxes(self) -> Path:
        return self._attach("_boxes.png")

    @property
    def final_boxes(self) -> Path:
        return self._attach("_boxes_final.png")

    @property
    def clean_json(self) -> Path:
        return self._attach("#clean.json")

    @property
    def box_mask(self) -> Path:
        return self._attach("_box_mask.png")

    @property
    def with_masks(self) -> Path:
        return self._attach("_with_masks.png")

    @property
    def cut_mask(self) -> Path:
        return self._attach("_cut_mask.png")

    @property
    def mask_fitments(self) -> Path:
        return self._attach("_mask_fitments.png")

    @property
    def combined_mask(self) -> Path:
        return self._attach("_combined_mask.png")

    @property
    def mask_overlay(self) -> Path:
        return self._attach("_with_masks.png")

    @property
    def std_devs(self) -> Path:
        return self._attach("_std_devs.png")

    @property
    def clean(self) -> Path:
        return self._attach("_clean.png")

    @property
    def text(self) -> Path:
        return self._attach("_text.png")

    @property
    def mask_data_json(self) -> Path:
        return self._attach("#mask_data.json")

    @property
    def noise_mask(self) -> Path:
        return self._attach("_noise_mask.png")

    @property
    def clean_denoised(self) -> Path:
        return self._attach("_clean_denoised.png")

    @property
    def inpainting(self) -> Path:
        return self._attach("_inpainting.png")

    @property
    def clean_inpaint(self) -> Path:
        return self._attach("_clean_inpaint.png")

    @property
    def psd(self) -> Path:
        return self._attach("_out.psd")

    @property
    def psd_bulk(self) -> Path:
        return self._attach("_out_bulk.psd")



import os
import subprocess

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger

import pcleaner.gui.post_action_config as pac
import pcleaner.gui.state_saver as ss
from pcleaner.gui.ui_generated_files.ui_PostActionRunner import Ui_PostActionRunner


class PostActionRunner(Qw.QDialog, Ui_PostActionRunner):
    """
    The post action execution dialog.
    """

    time_to_wait: int
    paused: bool
    dont_shut_down: bool
    state_saver: ss.StateSaver  # The state saver for the window.

    def __init__(
        self,
        parent=None,
        time_to_wait: int = 1,
        action_name: str = "",
        command: str = "",
        debug: bool = False,
        error: bool = False,
    ) -> None:
        """
        Init the widget.

        :param parent: The parent widget.
        """
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.paused = False
        self.time_to_wait = time_to_wait

        font = self.plainTextEdit_command.font()
        font.setFamily("Noto Mono")
        self.plainTextEdit_command.setFont(font)
        self.plainTextEdit_command.setPlainText(command)
        # Move the cursor to the end of the text.
        cursor = self.plainTextEdit_command.textCursor()
        cursor.movePosition(Qg.QTextCursor.End)
        self.plainTextEdit_command.setTextCursor(cursor)

        background_color = self.palette().color(Qg.QPalette.Window)
        theme_is_dark = background_color.lightness() < 128

        self.highlighter = ShellSyntaxHighlighter(
            self.plainTextEdit_command.document(), theme_is_dark
        )

        font = self.label_what.font()
        font.setPointSize(round(font.pointSize() * 1.5))
        self.label_what.setFont(font)

        if not error:
            self.label_error.hide()
            self.label_error_icon.hide()
        else:
            self.label_error_icon.setPixmap(Qg.QIcon.fromTheme("data-warning").pixmap(16, 16))

        self.dont_shut_down = action_name == pac.SHUTDOWN_COMMAND_NAME and debug

        if action_name == pac.SHUTDOWN_COMMAND_NAME:
            self.label_remaining_time.hide()
            self.pushButton_cancel_action.hide()
            self.label_action_icon.setPixmap(Qg.QIcon.fromTheme("system-shutdown").pixmap(64, 64))
            self.label_what.setText(self.tr("Shutting down the system"))
            self.lcdNumber_shutdown.setDigitCount(len(str(time_to_wait)))
            # I counted 18 pixels per digit.
            self.lcdNumber_shutdown.setFixedWidth(18 * len(str(time_to_wait)))
        else:
            self.lcdNumber_shutdown.hide()
            self.pushButton_cancel_shutdown.hide()
            self.label_action_icon.setPixmap(Qg.QIcon.fromTheme("dialog-scripts").pixmap(64, 64))
            self.label_what.setText(
                self.tr('Running the action "{action_name}"').format(action_name=action_name)
            )

        self.pushButton_again.hide()
        self.pushButton_again.clicked.connect(self.run_command)
        self.pushButton_close.hide()
        self.pushButton_close.clicked.connect(self.close)

        self.pushButton_pause.clicked.connect(self.pause)
        self.pushButton_resume.clicked.connect(self.resume)
        self.pushButton_resume.hide()
        self.pushButton_skip_countdown.clicked.connect(self.skip_countdown)

        self.pushButton_cancel_action.clicked.connect(self.cancel)
        self.pushButton_cancel_shutdown.clicked.connect(self.cancel)

        self.plainTextEdit_command.setTabStopDistance(40)

        if time_to_wait == 0:
            self.pushButton_again.show()
            self.pushButton_skip_countdown.hide()
            # Give the window a little time to appear.
            Qc.QTimer.singleShot(10, self.run_command)
        else:
            self.set_time_remaining(time_to_wait)
            self.count_down()

        self.state_saver = ss.StateSaver("post_action_runner")
        self.init_state_saver()
        self.state_saver.restore()

    def count_down(self) -> None:
        """
        Count down the time to wait.
        """
        if self.paused:
            return
        if self.time_to_wait > 0:
            if not self.paused:
                self.time_to_wait -= 1
                self.set_time_remaining(self.time_to_wait)
                Qc.QTimer.singleShot(1000, self.count_down)
        else:
            self.run_command()

    def run_command(self) -> None:
        if self.paused:
            # Last second cancellation.
            return
        if self.dont_shut_down:
            self.pushButton_again.show()
            self.pushButton_close.show()
            self.hide_countdown_stuff()
            logger.warning("Psyche! Not really shutting down, we're in debug mode.")
            return
        command = self.plainTextEdit_command.toPlainText()
        logger.info(f"Running command: {command}")
        subprocess.Popen(command, shell=True, start_new_session=True, env=os.environ)
        self.pushButton_again.show()
        self.pushButton_close.show()
        self.hide_countdown_stuff()

    def set_time_remaining(self, time: int) -> None:
        self.lcdNumber_shutdown.display(time)
        if time <= 3:
            # Make the number red.
            self.lcdNumber_shutdown.setStyleSheet("color: red")
        if time != 1:
            self.label_remaining_time.setText(
                self.tr("Remaining time: {time} seconds").format(time=time)
            )
        else:
            self.label_remaining_time.setText(self.tr("Remaining time: 1 second"))

    def init_state_saver(self) -> None:
        """
        Load the state from the state saver.
        """
        self.state_saver.register(self)

    def closeEvent(self, event: Qg.QCloseEvent) -> None:
        self.state_saver.save()
        self.paused = True
        event.accept()

    def pause(self) -> None:
        self.paused = True
        self.pushButton_pause.hide()
        self.pushButton_resume.show()

    def resume(self) -> None:
        self.paused = False
        self.pushButton_pause.show()
        self.pushButton_resume.hide()
        Qc.QTimer.singleShot(1000, self.count_down)

    def skip_countdown(self) -> None:
        self.time_to_wait = 0
        self.paused = False
        self.run_command()

    def cancel(self) -> None:
        self.paused = True
        self.pushButton_close.show()
        self.hide_countdown_stuff()

    def hide_countdown_stuff(self) -> None:
        self.pushButton_cancel_action.hide()
        self.pushButton_cancel_shutdown.hide()
        self.pushButton_pause.hide()
        self.pushButton_resume.hide()
        self.pushButton_skip_countdown.hide()
        self.label_remaining_time.hide()
        self.lcdNumber_shutdown.hide()

    def changeEvent(self, event) -> None:
        """
        Listen for palette change events to notify all widgets.
        """
        if event.type() == Qc.QEvent.PaletteChange:
            logger.debug("Palette changed, updating syntax highlighter.")
            background_color = self.palette().color(Qg.QPalette.Window)
            theme_is_dark = background_color.lightness() < 128
            self.highlighter.toggle_dark_mode(theme_is_dark)


class ShellSyntaxHighlighter(Qg.QSyntaxHighlighter):
    def __init__(self, document, dark_mode=False):
        super().__init__(document)

        self.dark_mode = dark_mode
        self.highlighting_rules = []

        self.color_map_light = {
            "comment": Qg.QColor(106, 153, 85),  # Greenish, subdued for comments
            "keyword": Qg.QColor(0, 81, 255),  # Bright blue for keywords
            "variable": Qg.QColor(75, 0, 130),  # Indigo for variables
            "string": Qg.QColor(163, 21, 21),  # Deep red for strings
            "number": Qg.QColor(60, 103, 64),  # Pastel green for numbers
            "operator": Qg.QColor(0, 122, 204),  # Soft blue for operators
            "path": Qg.QColor(36, 162, 36),  # Green for paths
        }

        self.color_map_dark = {
            "comment": Qg.QColor(106, 153, 85),  # Same greenish tone for comments
            "keyword": Qg.QColor(86, 156, 214),  # Soft cyan for keywords
            "variable": Qg.QColor(156, 220, 254),  # Bright cyan for variables
            "string": Qg.QColor(214, 157, 133),  # Soft coral for strings
            "number": Qg.QColor(181, 206, 168),  # Same pastel green for numbers
            "operator": Qg.QColor(212, 212, 212),  # Light gray for operators
            "path": Qg.QColor(255, 198, 109),  # Golden yellow for paths
        }
        self.setup_highlighting_rules()

    def setup_highlighting_rules(self):
        """Set up the highlighting rules based on the current color map."""
        color_map = self.color_map_dark if self.dark_mode else self.color_map_light
        self.highlighting_rules.clear()

        # Format for comments
        comment_format = Qg.QTextCharFormat()
        comment_format.setForeground(color_map["comment"])
        self.highlighting_rules.append((Qc.QRegularExpression(r"#.*"), comment_format))

        # Format for keywords
        keyword_format = Qg.QTextCharFormat()
        keyword_format.setForeground(color_map["keyword"])
        keywords = r"\b(if|then|else|fi|for|while|do|done|case|esac|in|function|elif|return|break|continue|export|unset|readonly)\b"
        self.highlighting_rules.append((Qc.QRegularExpression(keywords), keyword_format))

        # Format for variables
        variable_format = Qg.QTextCharFormat()
        variable_format.setForeground(color_map["variable"])
        self.highlighting_rules.append(
            (Qc.QRegularExpression(r"\$\w+|\$\{[^}]+\}"), variable_format)
        )

        # Format for strings
        string_format = Qg.QTextCharFormat()
        string_format.setForeground(color_map["string"])
        self.highlighting_rules.append(
            (Qc.QRegularExpression(r"'[^']*'|\"[^\"]*\""), string_format)
        )

        # Format for numbers
        number_format = Qg.QTextCharFormat()
        number_format.setForeground(color_map["number"])
        self.highlighting_rules.append((Qc.QRegularExpression(r"\b\d+\b"), number_format))

        # Format for operators and symbols
        operator_format = Qg.QTextCharFormat()
        operator_format.setForeground(color_map["operator"])
        self.highlighting_rules.append((Qc.QRegularExpression(r"[=|!<>;&]+"), operator_format))

        # File Paths (Linux/Mac)
        path_format = Qg.QTextCharFormat()
        path_format.setForeground(color_map["path"])
        self.highlighting_rules.append((Qc.QRegularExpression(r"(/[^ ]*|~[^ ]*)"), path_format))

        # Windows Absolute Paths
        win_path_format = Qg.QTextCharFormat()
        win_path_format.setForeground(color_map["path"])
        win_absolute = r"[A-Z]:\\(?:[^\\:*?\"<>|\r\n]+\\)*[^\\:*?\"<>|\r\n]*"
        self.highlighting_rules.append((Qc.QRegularExpression(win_absolute), win_path_format))

        # Windows Network Paths
        win_network = r"\\\\(?:[^\\:*?\"<>|\r\n]+\\)+[^\\:*?\"<>|\r\n]*"
        self.highlighting_rules.append((Qc.QRegularExpression(win_network), win_path_format))

        # Windows Escaped Paths
        win_escaped = r"[A-Z]:\\\\(?:[^\\:*?\"<>|\r\n]+\\\\)*[^\\:*?\"<>|\r\n]*"
        self.highlighting_rules.append((Qc.QRegularExpression(win_escaped), win_path_format))

    def toggle_dark_mode(self, dark_mode):
        """Toggle between light and dark mode and reapply highlighting rules."""
        self.dark_mode = dark_mode
        self.setup_highlighting_rules()
        self.rehighlight()

    def highlightBlock(self, text):
        """Apply syntax highlighting to a block of text."""
        for pattern, fmt in self.highlighting_rules:
            expression = pattern
            matches = expression.globalMatch(text)
            while matches.hasNext():
                match = matches.next()
                self.setFormat(match.capturedStart(), match.capturedLength(), fmt)



import re
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Signal


class RegexLineEdit(Qw.QWidget):
    """
    A little line edit combined with an indicator icon to show if the regex is valid or not.
    """

    def __init__(self):
        super().__init__()

        main_layout = Qw.QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)

        line_edit_layout = Qw.QHBoxLayout()
        self.line_edit = Qw.QLineEdit()
        self.line_edit.textChanged.connect(self.validate_regex)
        line_edit_layout.addWidget(self.line_edit)

        self.error_icon = Qw.QLabel()
        success_icon = Qg.QIcon.fromTheme("data-success")
        if not success_icon.isNull():
            self.error_icon.setPixmap(success_icon.pixmap(20, 20))  # Initially success icon
        self.error_icon.setToolTip("Regex is valid")
        line_edit_layout.addWidget(self.error_icon)
        line_edit_layout.setStretch(0, 1)  # Make the line edit stretchable
        line_edit_layout.setStretch(1, 0)  # Error icon should not stretch

        main_layout.addLayout(line_edit_layout)

        self.setLayout(main_layout)

    def validate_regex(self, text):
        try:
            re.compile(text)
            # Set success icon and tooltip
            success_icon = Qg.QIcon.fromTheme("data-success")
            if not success_icon.isNull():
                self.error_icon.setPixmap(success_icon.pixmap(20, 20))
            self.error_icon.setToolTip(self.tr("Regex is valid"))
        except re.error as e:
            # Set error icon from the theme and tooltip
            error_icon = Qg.QIcon.fromTheme("data-error")
            if not error_icon.isNull():
                self.error_icon.setPixmap(error_icon.pixmap(20, 20))
            self.error_icon.setToolTip(f"Regex Error: {e.msg}")

    def setRegex(self, text):
        self.line_edit.setText(text)

    def regex(self):
        return self.line_edit.text()

    @property
    def textChanged(self) -> Signal:
        return self.line_edit.textChanged

    def setClearButtonEnabled(self, enabled: bool):
        self.line_edit.setClearButtonEnabled(enabled)

    def setPlaceholderText(self, text: str):
        self.line_edit.setPlaceholderText(text)

    @property
    def textEdited(self):
        return self.line_edit.textEdited

    def text(self):
        return self.line_edit.text()

    def setText(self, text: str):
        self.line_edit.setText(text)

    def setRegexEnabled(self, enabled: bool):
        self.error_icon.setVisible(enabled)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'PostActionConfiguration.ui'
##
## Created by: Qt User Interface Compiler version 6.8.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QAbstractItemView, QAbstractSpinBox, QApplication,
    QCheckBox, QDialog, QDialogButtonBox, QFormLayout,
    QHBoxLayout, QHeaderView, QLabel, QLineEdit,
    QPushButton, QSizePolicy, QSpacerItem, QSpinBox,
    QTableWidgetItem, QVBoxLayout, QWidget)

from pcleaner.gui.CustomQ.CTableWidget import CTableWidget
from pcleaner.gui.CustomQ.CTooltipLabel import CTooltipLabel

class Ui_PostActionConfiguration(object):
    def setupUi(self, PostActionConfiguration):
        if not PostActionConfiguration.objectName():
            PostActionConfiguration.setObjectName(u"PostActionConfiguration")
        PostActionConfiguration.resize(800, 640)
        self.verticalLayout = QVBoxLayout(PostActionConfiguration)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.checkBox_remember_action = QCheckBox(PostActionConfiguration)
        self.checkBox_remember_action.setObjectName(u"checkBox_remember_action")

        self.verticalLayout.addWidget(self.checkBox_remember_action)

        self.horizontalLayout_8 = QHBoxLayout()
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.label_13 = QLabel(PostActionConfiguration)
        self.label_13.setObjectName(u"label_13")

        self.horizontalLayout_8.addWidget(self.label_13)

        self.spinBox_wait_time = QSpinBox(PostActionConfiguration)
        self.spinBox_wait_time.setObjectName(u"spinBox_wait_time")
        self.spinBox_wait_time.setMaximum(86400)
        self.spinBox_wait_time.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)

        self.horizontalLayout_8.addWidget(self.spinBox_wait_time)

        self.label_review_output_help_7 = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help_7.setObjectName(u"label_review_output_help_7")
        self.label_review_output_help_7.setToolTipDuration(-1)
        self.label_review_output_help_7.setText(u"<helper>")

        self.horizontalLayout_8.addWidget(self.label_review_output_help_7)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer_2)


        self.verticalLayout.addLayout(self.horizontalLayout_8)

        self.verticalSpacer_2 = QSpacerItem(20, 18, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setContentsMargins(-1, 0, -1, 6)
        self.label = QLabel(PostActionConfiguration)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label)

        self.lineEdit_shutdown = QLineEdit(PostActionConfiguration)
        self.lineEdit_shutdown.setObjectName(u"lineEdit_shutdown")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.lineEdit_shutdown)


        self.verticalLayout.addLayout(self.formLayout)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_2 = QLabel(PostActionConfiguration)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.pushButton_row_up = QPushButton(PostActionConfiguration)
        self.pushButton_row_up.setObjectName(u"pushButton_row_up")
        self.pushButton_row_up.setText(u"")
        icon = QIcon()
        iconThemeName = u"arrow-up"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_row_up.setIcon(icon)
        self.pushButton_row_up.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_row_up)

        self.pushButton_row_down = QPushButton(PostActionConfiguration)
        self.pushButton_row_down.setObjectName(u"pushButton_row_down")
        self.pushButton_row_down.setText(u"")
        icon1 = QIcon()
        iconThemeName = u"arrow-down"
        if QIcon.hasThemeIcon(iconThemeName):
            icon1 = QIcon.fromTheme(iconThemeName)
        else:
            icon1.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_row_down.setIcon(icon1)
        self.pushButton_row_down.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_row_down)

        self.pushButton_new = QPushButton(PostActionConfiguration)
        self.pushButton_new.setObjectName(u"pushButton_new")
        icon2 = QIcon()
        iconThemeName = u"list-add"
        if QIcon.hasThemeIcon(iconThemeName):
            icon2 = QIcon.fromTheme(iconThemeName)
        else:
            icon2.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_new.setIcon(icon2)
        self.pushButton_new.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_new)

        self.pushButton_delete = QPushButton(PostActionConfiguration)
        self.pushButton_delete.setObjectName(u"pushButton_delete")
        icon3 = QIcon()
        iconThemeName = u"edit-delete"
        if QIcon.hasThemeIcon(iconThemeName):
            icon3 = QIcon.fromTheme(iconThemeName)
        else:
            icon3.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_delete.setIcon(icon3)
        self.pushButton_delete.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_delete)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.tableWidget_commands = CTableWidget(PostActionConfiguration)
        if (self.tableWidget_commands.columnCount() < 2):
            self.tableWidget_commands.setColumnCount(2)
        __qtablewidgetitem = QTableWidgetItem()
        self.tableWidget_commands.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.tableWidget_commands.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        self.tableWidget_commands.setObjectName(u"tableWidget_commands")
        self.tableWidget_commands.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.tableWidget_commands.setProperty(u"showDropIndicator", False)
        self.tableWidget_commands.setDragDropOverwriteMode(False)
        self.tableWidget_commands.setAlternatingRowColors(True)
        self.tableWidget_commands.setSelectionMode(QAbstractItemView.SingleSelection)
        self.tableWidget_commands.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.tableWidget_commands.horizontalHeader().setDefaultSectionSize(200)
        self.tableWidget_commands.horizontalHeader().setStretchLastSection(True)

        self.verticalLayout.addWidget(self.tableWidget_commands)

        self.horizontalLayout_10 = QHBoxLayout()
        self.horizontalLayout_10.setObjectName(u"horizontalLayout_10")
        self.checkBox_cancel_on_error = QCheckBox(PostActionConfiguration)
        self.checkBox_cancel_on_error.setObjectName(u"checkBox_cancel_on_error")
        self.checkBox_cancel_on_error.setChecked(False)

        self.horizontalLayout_10.addWidget(self.checkBox_cancel_on_error)

        self.label_review_output_help_6 = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help_6.setObjectName(u"label_review_output_help_6")
        self.label_review_output_help_6.setToolTipDuration(-1)
        self.label_review_output_help_6.setText(u"<helper>")

        self.horizontalLayout_10.addWidget(self.label_review_output_help_6)

        self.horizontalSpacer_9 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_10.addItem(self.horizontalSpacer_9)


        self.verticalLayout.addLayout(self.horizontalLayout_10)

        self.verticalSpacer = QSpacerItem(20, 18, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.label_3 = QLabel(PostActionConfiguration)
        self.label_3.setObjectName(u"label_3")
        font = QFont()
        font.setBold(True)
        self.label_3.setFont(font)

        self.verticalLayout.addWidget(self.label_3)

        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setSpacing(24)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalLayout_7.setContentsMargins(-1, 6, -1, 6)
        self.formLayout_2 = QFormLayout()
        self.formLayout_2.setObjectName(u"formLayout_2")
        self.label_4 = QLabel(PostActionConfiguration)
        self.label_4.setObjectName(u"label_4")

        self.formLayout_2.setWidget(0, QFormLayout.LabelRole, self.label_4)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_5 = QLabel(PostActionConfiguration)
        self.label_5.setObjectName(u"label_5")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy)

        self.horizontalLayout_2.addWidget(self.label_5)

        self.label_review_output_help = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help.setObjectName(u"label_review_output_help")
        self.label_review_output_help.setToolTipDuration(-1)
        self.label_review_output_help.setText(u"<helper>")

        self.horizontalLayout_2.addWidget(self.label_review_output_help)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_3)


        self.formLayout_2.setLayout(0, QFormLayout.FieldRole, self.horizontalLayout_2)

        self.inputDirectoryIesLabel = QLabel(PostActionConfiguration)
        self.inputDirectoryIesLabel.setObjectName(u"inputDirectoryIesLabel")

        self.formLayout_2.setWidget(1, QFormLayout.LabelRole, self.inputDirectoryIesLabel)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_12 = QLabel(PostActionConfiguration)
        self.label_12.setObjectName(u"label_12")
        sizePolicy.setHeightForWidth(self.label_12.sizePolicy().hasHeightForWidth())
        self.label_12.setSizePolicy(sizePolicy)

        self.horizontalLayout_3.addWidget(self.label_12)

        self.label_review_output_help_2 = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help_2.setObjectName(u"label_review_output_help_2")
        self.label_review_output_help_2.setToolTipDuration(-1)
        self.label_review_output_help_2.setText(u"<helper>")

        self.horizontalLayout_3.addWidget(self.label_review_output_help_2)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_4)


        self.formLayout_2.setLayout(1, QFormLayout.FieldRole, self.horizontalLayout_3)


        self.horizontalLayout_7.addLayout(self.formLayout_2)

        self.formLayout_3 = QFormLayout()
        self.formLayout_3.setObjectName(u"formLayout_3")
        self.label_6 = QLabel(PostActionConfiguration)
        self.label_6.setObjectName(u"label_6")

        self.formLayout_3.setWidget(0, QFormLayout.LabelRole, self.label_6)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_9 = QLabel(PostActionConfiguration)
        self.label_9.setObjectName(u"label_9")
        sizePolicy.setHeightForWidth(self.label_9.sizePolicy().hasHeightForWidth())
        self.label_9.setSizePolicy(sizePolicy)

        self.horizontalLayout_4.addWidget(self.label_9)

        self.label_review_output_help_4 = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help_4.setObjectName(u"label_review_output_help_4")
        self.label_review_output_help_4.setToolTipDuration(-1)
        self.label_review_output_help_4.setText(u"<helper>")

        self.horizontalLayout_4.addWidget(self.label_review_output_help_4)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_5)


        self.formLayout_3.setLayout(0, QFormLayout.FieldRole, self.horizontalLayout_4)

        self.label_10 = QLabel(PostActionConfiguration)
        self.label_10.setObjectName(u"label_10")

        self.formLayout_3.setWidget(1, QFormLayout.LabelRole, self.label_10)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label_11 = QLabel(PostActionConfiguration)
        self.label_11.setObjectName(u"label_11")
        sizePolicy.setHeightForWidth(self.label_11.sizePolicy().hasHeightForWidth())
        self.label_11.setSizePolicy(sizePolicy)

        self.horizontalLayout_5.addWidget(self.label_11)

        self.label_review_output_help_5 = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help_5.setObjectName(u"label_review_output_help_5")
        self.label_review_output_help_5.setToolTipDuration(-1)
        self.label_review_output_help_5.setText(u"<helper>")

        self.horizontalLayout_5.addWidget(self.label_review_output_help_5)

        self.horizontalSpacer_6 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_6)


        self.formLayout_3.setLayout(1, QFormLayout.FieldRole, self.horizontalLayout_5)


        self.horizontalLayout_7.addLayout(self.formLayout_3)

        self.formLayout_4 = QFormLayout()
        self.formLayout_4.setObjectName(u"formLayout_4")
        self.label_8 = QLabel(PostActionConfiguration)
        self.label_8.setObjectName(u"label_8")

        self.formLayout_4.setWidget(0, QFormLayout.LabelRole, self.label_8)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.label_7 = QLabel(PostActionConfiguration)
        self.label_7.setObjectName(u"label_7")
        sizePolicy.setHeightForWidth(self.label_7.sizePolicy().hasHeightForWidth())
        self.label_7.setSizePolicy(sizePolicy)

        self.horizontalLayout_6.addWidget(self.label_7)

        self.label_review_output_help_3 = CTooltipLabel(PostActionConfiguration)
        self.label_review_output_help_3.setObjectName(u"label_review_output_help_3")
        self.label_review_output_help_3.setToolTipDuration(-1)
        self.label_review_output_help_3.setText(u"<helper>")

        self.horizontalLayout_6.addWidget(self.label_review_output_help_3)

        self.horizontalSpacer_7 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_7)


        self.formLayout_4.setLayout(0, QFormLayout.FieldRole, self.horizontalLayout_6)


        self.horizontalLayout_7.addLayout(self.formLayout_4)


        self.verticalLayout.addLayout(self.horizontalLayout_7)

        self.horizontalLayout_9 = QHBoxLayout()
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.label_warning_icon = QLabel(PostActionConfiguration)
        self.label_warning_icon.setObjectName(u"label_warning_icon")
        self.label_warning_icon.setText(u"<warning icon>")

        self.horizontalLayout_9.addWidget(self.label_warning_icon)

        self.label_warning = QLabel(PostActionConfiguration)
        self.label_warning.setObjectName(u"label_warning")
        font1 = QFont()
        font1.setItalic(True)
        self.label_warning.setFont(font1)
        self.label_warning.setText(u"<warning>")

        self.horizontalLayout_9.addWidget(self.label_warning)

        self.horizontalSpacer_8 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_9.addItem(self.horizontalSpacer_8)

        self.buttonBox = QDialogButtonBox(PostActionConfiguration)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.horizontalLayout_9.addWidget(self.buttonBox)

        self.horizontalLayout_9.setStretch(2, 1)

        self.verticalLayout.addLayout(self.horizontalLayout_9)


        self.retranslateUi(PostActionConfiguration)
        self.buttonBox.accepted.connect(PostActionConfiguration.accept)
        self.buttonBox.rejected.connect(PostActionConfiguration.reject)

        QMetaObject.connectSlotsByName(PostActionConfiguration)
    # setupUi

    def retranslateUi(self, PostActionConfiguration):
        PostActionConfiguration.setWindowTitle(QCoreApplication.translate("PostActionConfiguration", u"Configure Post-Run Actions", None))
        self.checkBox_remember_action.setText(QCoreApplication.translate("PostActionConfiguration", u"Always perform the action after processing", None))
        self.label_13.setText(QCoreApplication.translate("PostActionConfiguration", u"Wait before performing an action:", None))
        self.spinBox_wait_time.setSuffix(QCoreApplication.translate("PostActionConfiguration", u" s", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help_7.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>If you set a wait time greater than 0, a window will open upon process completion. This window will show a countdown and allow you to preview as well as edit the command.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label.setText(QCoreApplication.translate("PostActionConfiguration", u"Power Off Command:", None))
        self.lineEdit_shutdown.setPlaceholderText(QCoreApplication.translate("PostActionConfiguration", u"shutdown", None))
        self.label_2.setText(QCoreApplication.translate("PostActionConfiguration", u"Custom Commands:", None))
#if QT_CONFIG(tooltip)
        self.pushButton_row_up.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"Move command up", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_row_down.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"Move command down", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_new.setText(QCoreApplication.translate("PostActionConfiguration", u"New", None))
        self.pushButton_delete.setText(QCoreApplication.translate("PostActionConfiguration", u"Delete", None))
        ___qtablewidgetitem = self.tableWidget_commands.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("PostActionConfiguration", u"Action Name", None));
        ___qtablewidgetitem1 = self.tableWidget_commands.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("PostActionConfiguration", u"Command or Script File Path", None));
        self.checkBox_cancel_on_error.setText(QCoreApplication.translate("PostActionConfiguration", u"Cancel the custom action if processing failed", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help_6.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>The Power Off action will run regardless of this setting.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_3.setText(QCoreApplication.translate("PostActionConfiguration", u"Placeholder Tokens:", None))
        self.label_4.setText(QCoreApplication.translate("PostActionConfiguration", u"Input Files", None))
        self.label_5.setText(QCoreApplication.translate("PostActionConfiguration", u"%i", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>The individual file paths of each file processed, as a space separated list, with quotation marks where necessary.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.inputDirectoryIesLabel.setText(QCoreApplication.translate("PostActionConfiguration", u"Input Directory", None))
        self.label_12.setText(QCoreApplication.translate("PostActionConfiguration", u"%id", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help_2.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>Group files by parent directory and list these, as a space separated list, with quotation marks where necessary.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_6.setText(QCoreApplication.translate("PostActionConfiguration", u"Output Files", None))
        self.label_9.setText(QCoreApplication.translate("PostActionConfiguration", u"%o", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help_4.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>The individual file paths of each file created as an output, as a space separated list, with quotation marks where necessary.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_10.setText(QCoreApplication.translate("PostActionConfiguration", u"Output Directory", None))
        self.label_11.setText(QCoreApplication.translate("PostActionConfiguration", u"%od", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help_5.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>Group output files by parent directory and list these, as a space separated list, with quotation marks where necessary.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_8.setText(QCoreApplication.translate("PostActionConfiguration", u"Profile Used", None))
        self.label_7.setText(QCoreApplication.translate("PostActionConfiguration", u"%p", None))
#if QT_CONFIG(tooltip)
        self.label_review_output_help_3.setToolTip(QCoreApplication.translate("PostActionConfiguration", u"<html><head/><body><p>The name of the profile used, with quotation marks if necessary.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
    # retranslateUi




import platform
from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Signal, Slot
from loguru import logger
import cv2
import numpy as np

import pcleaner.gui.image_viewer as iv


class SwipeViewer(iv.ImageViewer):
    """
    Display 2 images overlayed on top of each other, with a percentage to control the
    point horizontally at which the other image is shown instead.
    """

    image_left: Qw.QGraphicsPixmapItem | None
    image_right: Qw.QGraphicsPixmapItem | None
    raw_image_left: Qg.QImage | None
    raw_image_right: Qg.QImage | None
    swipe_position: int
    slider: Qw.QSlider | None

    def __init__(self, parent=None) -> None:
        super(SwipeViewer, self).__init__(parent)
        self.image_left = None
        self.image_right = None
        self.swipe_position = 0
        self.slider = None

        self.raw_image_left = None
        self.raw_image_right = None

    def set_images(self, left: Path, right: Path) -> None:
        """
        Accept 2 image paths and load them into the viewer.
        """
        self.scene.clear()

        # Store the left path as the loaded image path.
        self.loaded_image_path = left

        self.image_left = Qw.QGraphicsPixmapItem()
        self.image_right = Qw.QGraphicsPixmapItem()
        self.scene.addItem(self.image_left)
        self.scene.addItem(self.image_right)
        # Set the parent's image item for zoom purposes.
        self.image_item = self.image_left

        self.image_left.setOffset(0, 0)
        self.image_right.setOffset(0, 0)

        pixmap = self.load_cache_aware(left)
        self.raw_image_left = pixmap.toImage()
        self.image_left.setPixmap(pixmap)

        self.raw_image_right = iv.load_image_with_orientation(right)
        pixmap = Qg.QPixmap.fromImage(self.raw_image_right)
        self.image_right.setPixmap(pixmap)

        self.image_dimensions = (
            self.image_item.pixmap().size().width(),
            self.image_item.pixmap().size().height(),
        )

        self.reset_scene_rect()
        self.update_smoothing()

        # Load the pixel dimensions into the slider.
        # The slider holds the exact number of pixels to show on the left side.
        if not self.slider:
            logger.error("Slider not connected to swipe view!")
        else:
            self.slider.setMaximum(self.raw_image_left.width())
            # Start the slider in the middle.
            self.slider.setValue(self.raw_image_left.width() // 2)

        self.update_mask()

    def update_mask(self) -> None:
        """
        Update the mask position based on the percentage.
        """
        if not self.image_left or not self.image_right:
            return

        width = self.raw_image_left.width()
        height = self.raw_image_left.height()

        # Create a new QPixmap for the first image.
        img1_part = self.raw_image_left.copy(0, 0, self.swipe_position, height)
        self.image_left.setPixmap(Qg.QPixmap.fromImage(img1_part))

        # Create a new QPixmap for the second image.
        img2_part = self.raw_image_right.copy(
            self.swipe_position,
            0,
            width - self.swipe_position,
            height,
        )
        self.image_right.setPixmap(Qg.QPixmap.fromImage(img2_part))
        self.image_right.setPos(self.swipe_position, 0)

    def connect_slider(self, slider: Qw.QSlider) -> None:
        self.slider = slider
        slider.valueChanged.connect(self.handle_slider_change)

    @Slot(int)
    def handle_slider_change(self, value: int) -> None:
        """
        Handle the slider value change.
        """
        self.swipe_position = value
        self.update_mask()

    def drawBackground(self, painter, rect) -> None:
        """
        Ignore the direct parent's background drawing,
        going straight to the graphicsview's function.
        """
        super(iv.ImageViewer, self).drawBackground(painter, rect)

    def set_transformation_mode(self, mode: Qc.Qt.TransformationMode) -> None:
        self.image_left.setTransformationMode(mode)
        self.image_right.setTransformationMode(mode)

    def drawForeground(self, painter, rect) -> None:
        # Draw a vertical line at the swipe position
        if self.image_left and self.image_right:
            # Get highlight color from the palette.
            highlight_color = self.palette().highlight().color()
            painter.setPen(Qg.QPen(highlight_color, 2))
            painter.drawLine(
                self.swipe_position, 0, self.swipe_position, self.raw_image_left.height()
            )


class OnionViewer(iv.ImageViewer):
    """
    Display 2 images overlayed on top of each other, with a percentage to control the
    transparency of the top image.
    """

    lower_image: Qw.QGraphicsPixmapItem | None
    upper_image: Qw.QGraphicsPixmapItem | None
    alpha: float
    slider: Qw.QSlider | None

    def __init__(self, parent=None) -> None:
        super(OnionViewer, self).__init__(parent)
        self.lower_image = None
        self.upper_image = None
        self.alpha = 0.5
        self.slider = None

    def set_images(self, lower: Path, upper: Path) -> None:
        """
        Accept 2 image paths and load them into the viewer.
        """
        self.scene.clear()

        # Set the lower image as the canonical image path.
        self.loaded_image_path = lower

        self.lower_image = Qw.QGraphicsPixmapItem()
        self.upper_image = Qw.QGraphicsPixmapItem()
        self.scene.addItem(self.lower_image)
        self.scene.addItem(self.upper_image)
        # Set the parent's image item for zoom purposes.
        self.image_item = self.lower_image

        self.lower_image.setOffset(0, 0)
        self.upper_image.setOffset(0, 0)

        pixmap = self.load_cache_aware(lower)
        self.lower_image.setPixmap(pixmap)

        upper_image = iv.load_image_with_orientation(upper)
        pixmap = Qg.QPixmap.fromImage(upper_image)
        self.upper_image.setPixmap(pixmap)

        self.image_dimensions = (
            self.image_item.pixmap().size().width(),
            self.image_item.pixmap().size().height(),
        )

        self.reset_scene_rect()
        self.update_smoothing()
        self.update_alpha()

    def update_alpha(self) -> None:
        if self.lower_image and self.upper_image:
            self.upper_image.setOpacity(self.alpha)

    def connect_slider(self, slider: Qw.QSlider) -> None:
        # Make the slider a percentage and start it at 50%.
        self.slider = slider
        slider.setMaximum(100)
        slider.valueChanged.connect(self.handle_slider_change)
        slider.setValue(50)

    @Slot(int)
    def handle_slider_change(self, value: int) -> None:
        self.alpha = value / 100
        self.update_alpha()


class DifferenceViewer(iv.ImageViewer):
    """
    Display 2 images overlaid on top of each other with the difference filter applied.
    An additional slider controls the opacity of the overlay.
    """

    lower_image: Qw.QGraphicsPixmapItem | None
    upper_image: Qw.QGraphicsPixmapItem | None
    alpha: float
    slider: Qw.QSlider | None

    def __init__(self, parent=None) -> None:
        super(DifferenceViewer, self).__init__(parent)
        self.lower_image = None
        self.upper_image = None
        self.alpha = 0.0
        self.slider = None

    def set_images(self, lower: Path, upper: Path) -> None:
        """
        Accept 2 image paths and load them into the viewer, applying the difference filter.
        """
        self.scene.clear()

        # Set the lower image as the canonical image path.
        self.loaded_image_path = lower

        self.lower_image = Qw.QGraphicsPixmapItem()
        self.upper_image = Qw.QGraphicsPixmapItem()
        self.scene.addItem(self.lower_image)
        self.scene.addItem(self.upper_image)
        # Set the parent's image item for zoom purposes.
        self.image_item = self.lower_image

        self.lower_image.setOffset(0, 0)
        self.upper_image.setOffset(0, 0)

        lower_image = cv2_imread(str(lower))
        upper_image = cv2_imread(str(upper))

        # Apply the difference filter using OpenCV
        difference_image = cv2.absdiff(lower_image, upper_image)

        # Convert back to QPixmap for display
        lower_pixmap = self.cv2_to_pixmap(lower_image)
        upper_pixmap = self.cv2_to_pixmap(difference_image)

        self.lower_image.setPixmap(lower_pixmap)
        self.upper_image.setPixmap(upper_pixmap)

        self.image_dimensions = (
            self.image_item.pixmap().size().width(),
            self.image_item.pixmap().size().height(),
        )

        self.reset_scene_rect()
        self.update_smoothing()
        self.update_alpha()

    def cv2_to_pixmap(self, cv2_img) -> Qg.QPixmap:
        """
        Convert an OpenCV image to QPixmap.
        """
        height, width, channel = cv2_img.shape
        bytes_per_line = 3 * width
        q_img = Qg.QImage(
            cv2_img.data, width, height, bytes_per_line, Qg.QImage.Format_RGB888
        ).rgbSwapped()
        return Qg.QPixmap.fromImage(q_img)

    def update_alpha(self) -> None:
        if self.lower_image and self.upper_image:
            self.upper_image.setOpacity(self.alpha)

    def connect_slider(self, slider: Qw.QSlider) -> None:
        # Make the slider a percentage and start it at 85%.
        self.slider = slider
        slider.setMaximum(100)
        slider.valueChanged.connect(self.handle_slider_change)
        slider.setValue(85)

    @Slot(int)
    def handle_slider_change(self, value: int) -> None:
        self.alpha = value / 100
        self.update_alpha()


class OverlayViewer(iv.ImageViewer):
    """
    Display 2 images overlaid on top of each other where the upper image is recolored to a solid color and the opacity is controlled by a slider.
    """

    lower_image: Qw.QGraphicsPixmapItem | None
    upper_image: Qw.QGraphicsPixmapItem | None
    alpha: float
    slider: Qw.QSlider | None
    recolor: tuple[int, int, int, int]

    def __init__(self, parent=None) -> None:
        super(OverlayViewer, self).__init__(parent)
        self.lower_image = None
        self.upper_image = None
        self.alpha = 0.5
        self.slider = None
        self.recolor = (255, 0, 0, 0)  # Default recolor to red

    def set_images(
        self, lower: Path, masks: list[Path], mask_color: tuple[int, int, int, int]
    ) -> None:
        """
        Accept 2 image paths and load them into the viewer.
        """
        self.scene.clear()

        # Set the lower image as the canonical image path.
        self.loaded_image_path = lower

        self.lower_image = Qw.QGraphicsPixmapItem()
        self.upper_image = Qw.QGraphicsPixmapItem()
        self.scene.addItem(self.lower_image)
        self.scene.addItem(self.upper_image)
        self.recolor = mask_color
        # Set the parent's image item for zoom purposes.
        self.image_item = self.lower_image

        self.lower_image.setOffset(0, 0)
        self.upper_image.setOffset(0, 0)

        lower_pixmap = self.load_cache_aware(lower)
        self.lower_image.setPixmap(lower_pixmap)

        self.image_dimensions = (
            self.image_item.pixmap().size().width(),
            self.image_item.pixmap().size().height(),
        )
        self.reset_scene_rect()

        # Merge masks
        merged_mask = self.merge_masks(masks, self.image_dimensions)

        # Apply recolor to the merged mask while preserving the alpha channel
        recolored_image = self.recolor_image(merged_mask, self.recolor)

        # Convert OpenCV images to QPixmap
        upper_pixmap = self.cv2_to_pixmap(recolored_image)
        self.upper_image.setPixmap(upper_pixmap)

        self.update_smoothing()
        self.update_alpha()

    @staticmethod
    def merge_masks(masks: list[Path], target_size: tuple[int, int]) -> np.ndarray:
        """
        Merge multiple mask images into one by adding them together.

        :param masks: The paths to the masks to join.
        :param target_size: The size of the original image that the masks need to be scaled to.
        :return: The layered masks.
        """
        merged_mask = None

        for i, mask_path in enumerate(masks):
            mask_image = cv2_imread(
                str(mask_path), cv2.IMREAD_UNCHANGED
            )  # Ensure alpha channel is loaded.

            if merged_mask is None:
                # Initialize the merged mask with zeros with the same size and type as the first mask.
                merged_mask = np.zeros_like(mask_image, dtype=np.uint8)

            # Resize the merged mask if the current mask is larger.
            if mask_image.shape[:2] != merged_mask.shape[:2]:
                if (
                    mask_image.shape[0] > merged_mask.shape[0]
                    or mask_image.shape[1] > merged_mask.shape[1]
                ):
                    merged_mask = cv2.resize(
                        merged_mask,
                        (mask_image.shape[1], mask_image.shape[0]),
                        interpolation=cv2.INTER_NEAREST,
                    )

            # Add the current mask to the merged mask.
            merged_mask = cv2.add(merged_mask, mask_image)

        # Resize the merged mask to the target size.
        # The numpy array has dimensions height,width so we need
        # to flip the order of the target_size.
        target_size_hw = target_size[1], target_size[0]
        if merged_mask.shape[:2] != target_size_hw:
            merged_mask = cv2.resize(merged_mask, target_size_hw, interpolation=cv2.INTER_NEAREST)

        return merged_mask

    @staticmethod
    def recolor_image(image: np.ndarray, color: tuple) -> np.ndarray:
        """
        Recolor the given image to the specified solid color while preserving the alpha channel.
        """
        # OpenCV uses BGR instead of RGB.
        b, g, r, a = cv2.split(image)
        recolored_image = np.zeros_like(image)
        recolored_image[:, :, 0] = color[2]  # B channel
        recolored_image[:, :, 1] = color[1]  # G channel
        recolored_image[:, :, 2] = color[0]  # R channel
        recolored_image[:, :, 3] = a  # Alpha channel
        return recolored_image

    @staticmethod
    def cv2_to_pixmap(cv2_img: np.ndarray) -> Qg.QPixmap:
        """
        Convert an OpenCV image to QPixmap.
        """
        height, width, channel = cv2_img.shape
        bytes_per_line = channel * width
        q_img = Qg.QImage(
            cv2_img.data,
            width,
            height,
            bytes_per_line,
            Qg.QImage.Format_ARGB32 if channel == 4 else Qg.QImage.Format_RGB888,
        )
        return Qg.QPixmap.fromImage(q_img)

    def update_alpha(self) -> None:
        if self.lower_image and self.upper_image:
            self.upper_image.setOpacity(self.alpha)

    def connect_slider(self, slider: Qw.QSlider) -> None:
        # Make the slider a percentage and start it at 50%.
        self.slider = slider
        slider.setMaximum(100)
        slider.valueChanged.connect(self.handle_slider_change)
        slider.setValue(50)

    @Slot(int)
    def handle_slider_change(self, value: int) -> None:
        self.alpha = value / 100
        self.update_alpha()


def cv2_imread(image_path: str, read_type=cv2.IMREAD_COLOR):
    """
    Windows is a retarded operating system that can't handle unicode correctly, breaking the
    cv2 imread function. This is a stupid workaround for that. At least it is about the
    same speed, only slightly slower.
    """

    if platform.system() == "Windows":
        return cv2.imdecode(np.fromfile(image_path, dtype=np.uint8), read_type)
    else:
        return cv2.imread(image_path, read_type)



import json
from enum import Enum, auto
from importlib import resources
from pathlib import Path
from typing import Sequence

from PIL import Image, ImageDraw, ImageFont, ImageEnhance
from attrs import frozen, define
from loguru import logger

import pcleaner.config as cfg
import pcleaner.data
import pcleaner.ocr.supported_languages as osl
import pcleaner.helpers as hp


class BoxType(Enum):
    BOX = 0
    EXTENDED_BOX = 1
    MERGED_EXT_BOX = 2
    REFERENCE_BOX = 3


@frozen
class Box:
    x1: int
    y1: int
    x2: int
    y2: int

    # You can create a box from a tuple of (x1, y1, x2, y2) using the * operator:
    # Box(*box_tuple)

    @property
    def as_tuple(self) -> tuple[int, int, int, int]:
        return self.x1, self.y1, self.x2, self.y2

    @property
    def as_tuple_xywh(self) -> tuple[int, int, int, int]:
        # QRect expects (x1, y1, width, height)
        return self.x1, self.y1, self.x2 - self.x1, self.y2 - self.y1

    def __str__(self) -> str:
        """
        String representation of the box coordinates, basically unpacking the tuple.

        :returns: The box coordinates as a string.
        """
        return f"{self.x1},{self.y1},{self.x2},{self.y2}"

    def __contains__(self, point: tuple[int, int]) -> bool:
        """
        Check if a point is inside the box.

        :param point: The point to check.
        :returns: True if the point is inside the box, False otherwise.
        """
        x, y = point
        return self.x1 <= x <= self.x2 and self.y1 <= y <= self.y2

    @property
    def area(self) -> int:
        return (self.x2 - self.x1) * (self.y2 - self.y1)

    @property
    def center(self) -> tuple[int, int]:
        return (self.x1 + self.x2) // 2, (self.y1 + self.y2) // 2

    def merge(self, box: "Box") -> "Box":
        x_min = min(self.x1, box.x1)
        y_min = min(self.y1, box.y1)
        x_max = max(self.x2, box.x2)
        y_max = max(self.y2, box.y2)

        return Box(x_min, y_min, x_max, y_max)

    def overlaps(self, other: "Box", threshold: float) -> bool:
        """
        Check if this box overlaps with another box.
        Merge the boxes if more than this percentage (0-100) of the smaller box is covered by the larger box.

        :param other: The other box to check for overlap.
        :param threshold: The threshold for the overlap check.
        :return: True if the boxes overlap, False otherwise.
        """
        # Calculate the area of the intersection.
        x_overlap = max(0, min(self.x2, other.x2) - max(self.x1, other.x1))
        y_overlap = max(0, min(self.y2, other.y2) - max(self.y1, other.y1))
        intersection = x_overlap * y_overlap
        # Get the area of the smaller box. Ensure this can never be 0.
        smaller_area = min(self.area, other.area) or 1

        return intersection / smaller_area > (threshold / 100)

    def overlaps_center(self, other: "Box") -> bool:
        return self.center in other or other.center in self

    def pad(self, amount: int, canvas_size: tuple[int, int]) -> "Box":
        """
        Grow the box by amount pixels, respecting the canvas size.

        :param amount: The amount of pixels to grow the box by.
        :param canvas_size: The size of the canvas to respect.
        """
        x1_new = max(self.x1 - amount, 0)
        y1_new = max(self.y1 - amount, 0)
        x2_new = min(self.x2 + amount, canvas_size[0])
        y2_new = min(self.y2 + amount, canvas_size[1])

        return Box(x1_new, y1_new, x2_new, y2_new)

    def right_pad(self, amount: int, canvas_size: tuple[int, int]) -> "Box":
        """
        Right-pad the box by amount pixels, respecting the canvas size.

        :param amount: The amount of pixels to right-pad the box by.
        :param canvas_size: The size of the canvas to respect.
        """
        x2_new = min(self.x2 + amount, canvas_size[0])

        return Box(self.x1, self.y1, x2_new, self.y2)

    def scale(self, factor: float) -> "Box":
        """
        Scale the box by a factor.

        :param factor: The factor to scale the box by.
        :return: The scaled box.
        """
        x1_new = int(self.x1 * factor)
        y1_new = int(self.y1 * factor)
        x2_new = int(self.x2 * factor)
        y2_new = int(self.y2 * factor)

        return Box(x1_new, y1_new, x2_new, y2_new)

    def translate(self, x: int, y: int) -> "Box":
        """
        Translate the box by x and y.

        :param x: The x translation.
        :param y: The y translation.
        :return: The translated box.
        """
        return Box(self.x1 + x, self.y1 + y, self.x2 + x, self.y2 + y)


@define
class PageData:
    """
    This dataclass represents the json data generated by the ai model.
    It contains the image path, mask path, the boxes, the extended boxes,
    and the merged extended boxes.

    - boxes: The boxes generated by the ai model, only expanded slightly for extra padding.
      These are used to make a tight-fitting mask.
    - extended_boxes: The boxes expanded a lot and used for masking off potential false positives.
    - merged_extended_boxes: Overlapping extended boxes are merged into one box, to prevent
      conflicts of overlapping mask regions. The original extended boxes are still kept,
      to provide a more precise mask for the initial cut. These are used to cut out the mask when
      analyzing the fit.
    - reference_boxes: These are extensions of the merged extended boxes, to make sure the mask
      has room to grow in the base image when providing analysis. These are used to cut out the
      base image when analyzing the fit.

    Boxes are represented as tuples of (x1, y1, x2, y2), where (x1, y1) is the top left corner,
    and (x2, y2) is the bottom right corner.
    """

    image_path: str  # Path to the copied png.
    mask_path: str  # Path to the generated mask.png
    original_path: str  # Path to the original image. (used for relative output)
    scale: float  # The size of the original image relative to the png.
    box_language: list[osl.LanguageCode | None]
    boxes: list[Box]
    extended_boxes: list[Box]
    merged_extended_boxes: list[Box]
    reference_boxes: list[Box]
    _image_size: tuple[int, int] = (
        None  # Cache the image size, so we don't have to load the image every time.
    )

    @classmethod
    def from_json(cls, json_str: str) -> "PageData":
        """
        Load a previously dumped PageData object from a json file.

        :param json_str: The json string to load from.
        """
        json_data = json.loads(json_str)
        return cls(
            json_data["image_path"],
            json_data["mask_path"],
            json_data["original_path"],
            json_data["scale"],
            [osl.to_language_code(lang) for lang in json_data["box_language"]],
            [Box(*b) for b in json_data["boxes"]],
            [Box(*b) for b in json_data["extended_boxes"]],
            [Box(*b) for b in json_data["merged_extended_boxes"]],
            [Box(*b) for b in json_data["reference_boxes"]],
        )

    def to_json(self) -> str:
        """
        Dump the PageData object to a json string.
        We want to exclude the _image_size attribute, since it's not needed.
        Box classes need to be serialized as tuples.
        """
        data = {
            "image_path": self.image_path,
            "mask_path": self.mask_path,
            "original_path": self.original_path,
            "scale": self.scale,
            "box_language": [str(lang) for lang in self.box_language],
            "boxes": [b.as_tuple for b in self.boxes],
            "extended_boxes": [b.as_tuple for b in self.extended_boxes],
            "merged_extended_boxes": [b.as_tuple for b in self.merged_extended_boxes],
            "reference_boxes": [b.as_tuple for b in self.reference_boxes],
        }
        return json.dumps(data, indent=4)

    @property
    def image_size(self) -> tuple[int, int]:
        if self._image_size is None:
            self._image_size = Image.open(self.image_path).size
        return self._image_size

    def boxes_from_type(self, box_type: BoxType) -> list[Box]:
        match box_type:
            case BoxType.BOX:
                return self.boxes
            case BoxType.EXTENDED_BOX:
                return self.extended_boxes
            case BoxType.MERGED_EXT_BOX:
                return self.merged_extended_boxes
            case BoxType.REFERENCE_BOX:
                return self.reference_boxes
            case _:
                raise ValueError("Invalid box type.")

    def grow_boxes(self, padding: int, box_type: BoxType) -> None:
        """
        Uniformly grow all boxes by padding pixels.
        Checks that the boxes don't go out of bounds.

        :param padding: number of pixels to grow each box by.
        :param box_type: type of box to grow.
        """
        boxes = self.boxes_from_type(box_type)
        for i, box in enumerate(boxes):
            boxes[i] = box.pad(padding, self.image_size)

    def right_pad_boxes(self, padding: int, box_type: BoxType) -> None:
        """
        Right-pad all boxes by padding pixels.
        Checks that the boxes don't go out of bounds.

        :param padding: number of pixels to right-pad each box by.
        :param box_type: type of box to right-pad.
        """
        boxes = self.boxes_from_type(box_type)
        for i, box in enumerate(boxes):
            boxes[i] = box.right_pad(padding, self.image_size)

    def visualize(self, image_path: Path | str, output_path: Path | str) -> None:
        """
        Visualize the boxes on an image.
        Typically, this would be used to check where on the original image the
        boxes are located.

        :param image_path: The path to the image to visualize the boxes on.
        :param output_path: A full file path to write the image to.
        """
        image = Image.open(image_path)
        font_path = hp.resource_path(pcleaner.data, "LiberationSans-Regular.ttf")
        logger.debug(f"Loading included font from {font_path}")
        # Figure out the optimal font size based on the image size. E.g. 30 for a 1600px image.
        font_size = int(image.size[0] / 50) + 5

        # First, draw all the rectangles with a transparent fill.
        # This is done on a temporary image with full opacity to then
        # make the whole layer semi-transparent, without needing to
        # carefully fill the gaps between boxes.
        FILL_ALPHA = 48
        fill_layer = Image.new("RGBA", image.size, (0, 0, 0, 0))
        draw = ImageDraw.Draw(fill_layer)
        for box in self.reference_boxes:
            draw.rectangle(box.as_tuple, fill="blue")
        for box in self.merged_extended_boxes:
            draw.rectangle(box.as_tuple, fill=(128, 0, 200, 255))  # Better purple
        for box in self.extended_boxes:
            draw.rectangle(box.as_tuple, fill="red")
        for box in self.boxes:
            draw.rectangle(box.as_tuple, fill="green")

        # Apply transparency to the fill layer
        alpha = fill_layer.split()[3]
        alpha = ImageEnhance.Brightness(alpha).enhance(FILL_ALPHA / 255)
        fill_layer.putalpha(alpha)

        # Composite the fill layer onto the original image
        image = Image.alpha_composite(image.convert("RGBA"), fill_layer)
        draw = ImageDraw.Draw(image)

        for index, box in enumerate(self.boxes):
            draw.rectangle(box.as_tuple, outline="green")
            # Draw the box number, with a white background, respecting font size.
            draw.text(
                (box.x1 + 4, box.y1),
                str(index + 1),
                fill="green",
                font=ImageFont.truetype(font_path, font_size),
                stroke_fill="white",
                stroke_width=3,
            )

        for box in self.extended_boxes:
            draw.rectangle(box.as_tuple, outline="red")
        for box in self.merged_extended_boxes:
            draw.rectangle(box.as_tuple, outline="purple")
        for box in self.reference_boxes:
            draw.rectangle(box.as_tuple, outline="blue")

        image.save(output_path)

    def make_box_mask(self, image_size: tuple[int, int], box_type: BoxType) -> Image:
        """
        Draw the boxes as a bitmap mask, where 1 represents box, and 0 represents no box.
        This is in essence just an image where the color is either black or white.

        :param image_size: The size of the image to make the mask for.
        :param box_type: The type of box to use.
        :return: The mask. Image mode: "1"
        """
        box_mask = Image.new("1", image_size, (0,))
        draw = ImageDraw.Draw(box_mask)
        boxes = self.boxes_from_type(box_type)
        for box in boxes:
            draw.rectangle(box.as_tuple, fill=(1,))
        return box_mask

    def resolve_total_overlaps(self) -> None:
        """
        Check the initial boxes for overlaps where the center of either box is within the other box.
        This kind of overlap means that the same text is included in both boxes, they aren't merely touching
        at the edges. This way we don't get duplicate text in OCR output.
        """
        # Place the extended boxes in the merged extended boxes, merging overlapping boxes.
        merge_queue = self.boxes.copy()
        merged_boxes = []
        while merge_queue:
            box = merge_queue.pop(0)
            # Find all boxes that overlap with this box.
            overlapping_boxes = [b for b in merge_queue if box.overlaps_center(b)]
            # Merge all overlapping boxes.
            for b in overlapping_boxes:
                box = box.merge(b)
                merge_queue.remove(b)
            merged_boxes.append(box)

        self.boxes = merged_boxes

    def resolve_overlaps(self, from_type: BoxType, to_type: BoxType, threshold: float) -> None:
        """
        Copy the extended boxes to the merged extended boxes, and merge overlapping boxes.

        :param from_type: The type of boxes to merge.
        :param to_type: The type of boxes to overwrite with the merged boxes.
        :param threshold: The threshold for the overlap check. Merge if more than this percentage (0-100) of the smaller
            box is covered by the larger box.
        """
        # Place the extended boxes in the merged extended boxes, merging overlapping boxes.
        merge_queue = set(self.boxes_from_type(from_type))
        merged_boxes = []
        while merge_queue:
            box = merge_queue.pop()
            # Find all boxes that overlap with this box.
            overlapping_boxes = [b for b in merge_queue if box.overlaps(b, threshold)]
            # Merge all overlapping boxes.
            for b in overlapping_boxes:
                box = box.merge(b)
                merge_queue.remove(b)
            merged_boxes.append(box)

        boxes_reference = self.boxes_from_type(to_type)
        boxes_reference.clear()
        boxes_reference.extend(merged_boxes)


@frozen
class OCRAnalytic:
    """
    Analytics data to quantify the OCR performance and bundle output in a usable format.
    - number of boxes
    - sizes of all boxes that were ocred
    - sizes of the boxes that were removed
    - the cached file name and the text and the box that was removed.
    """

    path: Path
    num_boxes: int
    box_sizes_ocr: Sequence[int]
    box_sizes_removed: Sequence[int]
    removed_box_data: Sequence[tuple[str, Box]]


class OCRStatus(Enum):
    Normal = auto()
    Removed = auto()
    Edited = auto()
    EditedRemoved = auto()
    New = auto()


@define
class OCRResult:
    """
    A mutable variant that contains the OCR results for a single image.
    This is used in the OCR review window to allow human editing.

    The label contains the original index for process-created boxes,
    but newly created boxes are labeled with "New X", where X is the index of newly added boxes.

    bubbles: list[tuple[Path, str, Box, str, OCRStatus]]
    """

    path: Path
    text: str
    box: Box
    label: str
    status: OCRStatus


def convert_ocr_analytics_to_results(ocr_analytics: list[OCRAnalytic]) -> list[list[OCRResult]]:
    """
    We pretty much just extract the removed box data from the analytics, but attach a
    status to each box.

    :param ocr_analytics: The OCR analytics to convert.
    :return: The OCR results per image file.
    """
    ocr_results = []
    for analytic in ocr_analytics:
        bubbles = [
            OCRResult(analytic.path, text, box, str(index), OCRStatus.Normal)
            for index, (text, box) in enumerate(analytic.removed_box_data, start=1)
        ]
        ocr_results.append(bubbles)
    return ocr_results


def convert_ocr_results_to_analytics(ocr_results: list[list[OCRResult]]) -> list[OCRAnalytic]:
    """
    Just drop the status and leave the rest of the data blank,
    meaning only the removed box data is preserved.
    Boxes with the status Removed are dropped.

    :param ocr_results: The OCR results to convert.
    :return: The OCR analytics per image file.
    """
    ocr_analytics = []
    for results in ocr_results:
        keeping_box_data = [
            (result.text, result.box) for result in results if result.status != OCRStatus.Removed
        ]
        if not keeping_box_data:
            continue

        ocr_analytics.append(
            OCRAnalytic(
                results[0].path,
                len(keeping_box_data),
                [],
                [],
                keeping_box_data,
            )
        )
    return ocr_analytics


def merge_ocr_analytics(
    split_from: Path, segment_paths: list[Path], ocr_analytics: list[OCRAnalytic]
) -> None:
    """
    We need to merge and correct coordinates for the OCR analytics.
    OCRAnalytics are modified in place.

    :param split_from: The path to the original image.
    :param segment_paths: The paths to the split images.
    :param ocr_analytics: The full list of OCR analytics.
    """

    # Analytic:
    # num_boxes: int
    # box_sizes_ocr: Sequence[int]
    # box_sizes_removed: Sequence[int]
    # removed_box_data: Sequence[tuple[Path, str, Box]]

    # First we need to find which analytics to merge by identifying the
    # file name in one of the removed_box_data entries.
    # It can happen that no analytics are found, in which case we create a dummy one.
    split_file_to_analytic: dict[Path, OCRAnalytic] = {}
    for segment_path in segment_paths:
        for analytic in ocr_analytics:
            if segment_path == analytic.path:
                split_file_to_analytic[segment_path] = analytic
                break
        else:
            # Create a dummy analytic. This shouldn't happen, but this guards against future oopsies.
            logger.warning(f"Creating dummy analytic for {segment_path}")
            split_file_to_analytic[segment_path] = OCRAnalytic(segment_path, 0, (), (), ())

    # Purge the assigned analytics from the main list.
    for analytic_found in split_file_to_analytic.values():
        # We may have added dummies, so we need to not remove those.
        if analytic_found in ocr_analytics:
            ocr_analytics.remove(analytic_found)

    merge_num_boxes = sum(analytic.num_boxes for analytic in split_file_to_analytic.values())
    merge_box_sizes_ocr = [
        size for analytic in split_file_to_analytic.values() for size in analytic.box_sizes_ocr
    ]
    merge_box_sizes_removed = [
        size for analytic in split_file_to_analytic.values() for size in analytic.box_sizes_removed
    ]
    # For the boxes we need to adjust the coordinates and fix the file path.
    new_box_tuples = []
    offset = 0
    for segment_path in segment_paths:
        for text, box in split_file_to_analytic[segment_path].removed_box_data:
            new_box = box.translate(0, offset)
            new_box_tuples.append((text, new_box))
        offset += Image.open(segment_path).size[1]

    # Create the new analytic and add it to the list.
    new_analytic = OCRAnalytic(
        split_from,
        merge_num_boxes,
        merge_box_sizes_ocr,
        merge_box_sizes_removed,
        new_box_tuples,
    )
    ocr_analytics.append(new_analytic)


@frozen
class MaskFittingAnalytic:
    """
    Analytics data to visualize the mask fitting performance.
    - The image path.
    - Whether a good enough fit was found.
    - The mask index chosen by the mask fitting process.
    - The standard deviation of the mask chosen by the mask fitting process.
    - The thickness of the mask chosen by the mask fitting process. None if the box mask.
    """

    image_path: Path
    fit_was_found: bool
    mask_index: int
    mask_std_deviation: float
    mask_thickness: int | None


@frozen
class MaskFittingResults:
    """
    This is a simple struct to hold the results from the mask fitting process.
    Since it returns a lot of data, this is more readable.
    """

    best_mask: Image
    median_color: tuple[int, int, int]
    mask_coords: tuple[int, int]
    analytics_page_path: Path
    analytics_std_deviation: float
    analytics_mask_index: int
    analytics_thickness: int | None
    mask_box: Box  # Used for the denoising process.
    debug_masks: list[Image]

    @property
    def analytics(self) -> MaskFittingAnalytic:
        return MaskFittingAnalytic(
            self.analytics_page_path,
            not self.failed,
            self.analytics_mask_index,
            self.analytics_std_deviation,
            self.analytics_thickness,
        )

    @property
    def failed(self) -> bool:
        return self.best_mask is None

    @property
    def mask_data(self) -> tuple[Image, tuple[int, int, int], tuple[int, int]]:
        return self.best_mask, self.median_color, self.mask_coords

    @property
    def noise_mask_data(self) -> tuple[Box, float]:
        return self.mask_box, self.analytics_std_deviation


@frozen
class MaskerData:
    """
    This is a simple struct to hold the inputs for the masker.
    The data is a tuple of:
    - The json file path.
    - The image cache directory.
    - The masker config.
    - The extract text flag.
    - The show masks flag. (when true, save intermediate masks to the cache directory)
    - The debug flag.
    """

    json_path: Path
    cache_dir: Path
    masker_config: cfg.MaskerConfig
    extract_text: bool
    show_masks: bool
    debug: bool


@frozen
class MaskData:
    """
    This is a simple struct to hold all the extra information needed to perform the
    denoising process.

    - The original image path.
    - The cached image path.
    - The mask image path.
    - The scale of the original image to the base image.
    - The box coordinates with their respective standard deviation for the masks,
      whether they failed, and the mask thickness.
    """

    original_path: Path
    base_image_path: Path
    mask_path: Path
    scale: float
    boxes_with_stats: Sequence[tuple[Box, float, bool, int | None]]

    @classmethod
    def from_json(cls, json_str: str) -> "MaskData":
        """
        Create a MaskData object from a json string.
        Boxes with deviation need to be deserialized from tuples.

        :param json_str: The json string.
        :return: The MaskData object.
        """
        data = json.loads(json_str)
        return cls(
            Path(data["original_path"]),
            Path(data["base_image_path"]),
            Path(data["mask_path"]),
            data["scale"],
            [
                (Box(*box), deviation, failed, thickness)
                for box, deviation, failed, thickness in data["boxes_with_stats"]
            ],
        )

    def to_json(self) -> str:
        """
        Dump the MaskData object to a json string.
        Boxes are serialized as tuples.
        """
        # Convert the Path objects to strings.
        data = {
            "original_path": str(self.original_path),
            "base_image_path": str(self.base_image_path),
            "mask_path": str(self.mask_path),
            "scale": self.scale,
            "boxes_with_stats": [
                (box.as_tuple, deviation, failed, thickness)
                for box, deviation, failed, thickness in self.boxes_with_stats
            ],
        }
        return json.dumps(data, indent=4)


@frozen
class DenoiserData:
    """
    This is a simple struct to hold the inputs for the denoiser.
    The data is a tuple of:
    - The json file path.
    - The image cache directory.
    - The denoiser config.
    - The debug flag.
    """

    json_path: Path
    cache_dir: Path
    denoiser_config: cfg.DenoiserConfig
    debug: bool


@frozen
class DenoiseAnalytic:
    """
    Analytics data to visualize the denoising performance.
    - The standard deviations of the mask selection process. They are shown here,
      due to them being relevant to the min-threshold for choosing masks to denoise.
    - The path to the original image. This is needed to trace the analytics back to the image in the gui.
    """

    std_deviations: Sequence[float]
    path: Path


@frozen
class InpainterData:
    """
    This is a simple struct to hold the inputs for the inpainter.
    The data is a tuple of:
    - The page data json path. (This is #clean.json)
    - The mask data json path.  (This is #mask_data.json)
    - The image output directory.
    - The image cache directory.
    - The general config.
    - The masker config.  (For the min mask size)
    - The inpainter config.
    - The show masks flag. (when true, save intermediate masks to the cache directory)
    - The debug flag.
    """

    page_data_json_path: Path
    mask_data_json_path: Path
    cache_dir: Path
    masker_config: cfg.MaskerConfig
    denoiser_config: cfg.DenoiserConfig
    inpainter_config: cfg.InpainterConfig
    debug: bool


@frozen
class InpaintingAnalytic:
    """
    Analytics data to visualize the inpainting performance.
    - The thickness of the outline when inpainting.
    - The path to the original image.
    """

    thicknesses: Sequence[int]
    path: Path



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'PostActionRunner.ui'
##
## Created by: Qt User Interface Compiler version 6.8.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QFrame, QHBoxLayout,
    QLCDNumber, QLabel, QPlainTextEdit, QPushButton,
    QSizePolicy, QSpacerItem, QVBoxLayout, QWidget)

class Ui_PostActionRunner(object):
    def setupUi(self, PostActionRunner):
        if not PostActionRunner.objectName():
            PostActionRunner.setObjectName(u"PostActionRunner")
        PostActionRunner.resize(600, 400)
        self.verticalLayout_3 = QVBoxLayout(PostActionRunner)
        self.verticalLayout_3.setSpacing(12)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_action_icon = QLabel(PostActionRunner)
        self.label_action_icon.setObjectName(u"label_action_icon")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_action_icon.sizePolicy().hasHeightForWidth())
        self.label_action_icon.setSizePolicy(sizePolicy)
        self.label_action_icon.setMinimumSize(QSize(64, 64))
        self.label_action_icon.setText(u"<icon>")

        self.horizontalLayout_2.addWidget(self.label_action_icon)

        self.verticalLayout_4 = QVBoxLayout()
        self.verticalLayout_4.setSpacing(4)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(-1, -1, 18, -1)
        self.verticalSpacer_2 = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer_2)

        self.label_what = QLabel(PostActionRunner)
        self.label_what.setObjectName(u"label_what")
        self.label_what.setText(u"<doing what>")

        self.verticalLayout_4.addWidget(self.label_what)

        self.label_remaining_time = QLabel(PostActionRunner)
        self.label_remaining_time.setObjectName(u"label_remaining_time")
        self.label_remaining_time.setText(u"<remaining time>")

        self.verticalLayout_4.addWidget(self.label_remaining_time)

        self.verticalSpacer_3 = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer_3)


        self.horizontalLayout_2.addLayout(self.verticalLayout_4)

        self.lcdNumber_shutdown = QLCDNumber(PostActionRunner)
        self.lcdNumber_shutdown.setObjectName(u"lcdNumber_shutdown")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.lcdNumber_shutdown.sizePolicy().hasHeightForWidth())
        self.lcdNumber_shutdown.setSizePolicy(sizePolicy1)
        self.lcdNumber_shutdown.setMinimumSize(QSize(0, 48))
        self.lcdNumber_shutdown.setFrameShape(QFrame.NoFrame)
        self.lcdNumber_shutdown.setDigitCount(5)
        self.lcdNumber_shutdown.setSegmentStyle(QLCDNumber.Flat)

        self.horizontalLayout_2.addWidget(self.lcdNumber_shutdown)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.horizontalLayout_2.setStretch(3, 6)

        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setSpacing(4)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_error_icon = QLabel(PostActionRunner)
        self.label_error_icon.setObjectName(u"label_error_icon")
        sizePolicy.setHeightForWidth(self.label_error_icon.sizePolicy().hasHeightForWidth())
        self.label_error_icon.setSizePolicy(sizePolicy)
        self.label_error_icon.setMinimumSize(QSize(16, 16))
        self.label_error_icon.setText(u"<icon>")

        self.horizontalLayout_4.addWidget(self.label_error_icon)

        self.label_error = QLabel(PostActionRunner)
        self.label_error.setObjectName(u"label_error")
        font = QFont()
        font.setItalic(True)
        self.label_error.setFont(font)

        self.horizontalLayout_4.addWidget(self.label_error)


        self.verticalLayout_3.addLayout(self.horizontalLayout_4)

        self.verticalSpacer = QSpacerItem(20, 6, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

        self.verticalLayout_3.addItem(self.verticalSpacer)

        self.label_2 = QLabel(PostActionRunner)
        self.label_2.setObjectName(u"label_2")
        font1 = QFont()
        font1.setBold(True)
        self.label_2.setFont(font1)

        self.verticalLayout_3.addWidget(self.label_2)

        self.plainTextEdit_command = QPlainTextEdit(PostActionRunner)
        self.plainTextEdit_command.setObjectName(u"plainTextEdit_command")
        self.plainTextEdit_command.setPlainText(u"")
        self.plainTextEdit_command.setTextInteractionFlags(Qt.TextEditorInteraction)

        self.verticalLayout_3.addWidget(self.plainTextEdit_command)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setSpacing(6)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalSpacer_3 = QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_3)

        self.verticalLayout_5 = QVBoxLayout()
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.pushButton_again = QPushButton(PostActionRunner)
        self.pushButton_again.setObjectName(u"pushButton_again")
        icon = QIcon(QIcon.fromTheme(u"debug-run"))
        self.pushButton_again.setIcon(icon)

        self.verticalLayout_5.addWidget(self.pushButton_again)

        self.pushButton_skip_countdown = QPushButton(PostActionRunner)
        self.pushButton_skip_countdown.setObjectName(u"pushButton_skip_countdown")
        icon1 = QIcon(QIcon.fromTheme(u"media-skip-forward"))
        self.pushButton_skip_countdown.setIcon(icon1)

        self.verticalLayout_5.addWidget(self.pushButton_skip_countdown)


        self.horizontalLayout_3.addLayout(self.verticalLayout_5)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.pushButton_resume = QPushButton(PostActionRunner)
        self.pushButton_resume.setObjectName(u"pushButton_resume")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.pushButton_resume.sizePolicy().hasHeightForWidth())
        self.pushButton_resume.setSizePolicy(sizePolicy2)
        icon2 = QIcon(QIcon.fromTheme(u"media-playback-start"))
        self.pushButton_resume.setIcon(icon2)

        self.verticalLayout.addWidget(self.pushButton_resume)

        self.pushButton_pause = QPushButton(PostActionRunner)
        self.pushButton_pause.setObjectName(u"pushButton_pause")
        icon3 = QIcon(QIcon.fromTheme(u"media-playback-pause"))
        self.pushButton_pause.setIcon(icon3)

        self.verticalLayout.addWidget(self.pushButton_pause)


        self.horizontalLayout_3.addLayout(self.verticalLayout)

        self.pushButton_close = QPushButton(PostActionRunner)
        self.pushButton_close.setObjectName(u"pushButton_close")
        icon4 = QIcon(QIcon.fromTheme(u"window-close"))
        self.pushButton_close.setIcon(icon4)

        self.horizontalLayout_3.addWidget(self.pushButton_close)

        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.pushButton_cancel_action = QPushButton(PostActionRunner)
        self.pushButton_cancel_action.setObjectName(u"pushButton_cancel_action")
        icon5 = QIcon(QIcon.fromTheme(u"dialog-cancel"))
        self.pushButton_cancel_action.setIcon(icon5)

        self.verticalLayout_2.addWidget(self.pushButton_cancel_action)

        self.pushButton_cancel_shutdown = QPushButton(PostActionRunner)
        self.pushButton_cancel_shutdown.setObjectName(u"pushButton_cancel_shutdown")
        icon6 = QIcon()
        iconThemeName = u"dialog-cancel"
        if QIcon.hasThemeIcon(iconThemeName):
            icon6 = QIcon.fromTheme(iconThemeName)
        else:
            icon6.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_cancel_shutdown.setIcon(icon6)

        self.verticalLayout_2.addWidget(self.pushButton_cancel_shutdown)


        self.horizontalLayout_3.addLayout(self.verticalLayout_2)


        self.verticalLayout_3.addLayout(self.horizontalLayout_3)

        self.verticalLayout_3.setStretch(4, 1)

        self.retranslateUi(PostActionRunner)

        QMetaObject.connectSlotsByName(PostActionRunner)
    # setupUi

    def retranslateUi(self, PostActionRunner):
        PostActionRunner.setWindowTitle(QCoreApplication.translate("PostActionRunner", u"Post-Action Runner", None))
        self.label_error.setText(QCoreApplication.translate("PostActionRunner", u"An error occurred during the process.", None))
        self.label_2.setText(QCoreApplication.translate("PostActionRunner", u"Full Command:", None))
        self.pushButton_again.setText(QCoreApplication.translate("PostActionRunner", u"Run Again", None))
        self.pushButton_skip_countdown.setText(QCoreApplication.translate("PostActionRunner", u"Skip Countdown", None))
        self.pushButton_resume.setText(QCoreApplication.translate("PostActionRunner", u"Resume Countdown", None))
        self.pushButton_pause.setText(QCoreApplication.translate("PostActionRunner", u"Pause Countdown", None))
        self.pushButton_close.setText(QCoreApplication.translate("PostActionRunner", u"Close", None))
        self.pushButton_cancel_action.setText(QCoreApplication.translate("PostActionRunner", u"Cancel Action", None))
        self.pushButton_cancel_shutdown.setText(QCoreApplication.translate("PostActionRunner", u"Cancel Shutdown", None))
    # retranslateUi




import os
from copy import deepcopy, copy
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Callable

import torch
from loguru import logger

import pcleaner.config as cfg
import pcleaner.denoiser as dn
import pcleaner.gui.ctd_interface_gui as ctm
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.worker_thread as wt
import pcleaner.gui.structures as gst
import pcleaner.image_export as ie
import pcleaner.inpainting as ip
import pcleaner.masker as ma
import pcleaner.ocr.ocr as ocr
import pcleaner.output_structures as ost
import pcleaner.preprocessor as pp
import pcleaner.structures as st
import pcleaner.helpers as hp
from pcleaner import model_downloader as md
from pcleaner.config import LayeredExport
from pcleaner.helpers import tr


def generate_output(
    image_objects: list[imf.ImageFile | imf.MergedImageFile],
    split_files: dict[Path, list[imf.ImageFile]],
    target_outputs: list[ost.Output],
    output_dir: Path | None,
    config: cfg.Config,
    ocr_processor: ocr.OCREngineFactory | None,
    progress_callback: ost.ProgressSignal,
    batch_metadata: gst.BatchMetadata,
    abort_flag: wt.SharableFlag,
    debug: bool = False,
):
    """
    Process all the images in the given image objects list until the target step is reached.
    The paths inside the image objects are updated as the processing progresses.
    If the profile checksum of a step hasn't changed, the step is skipped.

    In GUI mode, the masker and denoiser never write the output directly, but instead
    place them in the cache directory. From there they may be copied to the output directory,
    to ensure that effort isn't duplicated where possible. For example, if the output was already
    generated for the viewer, it doesn't need to be generated again for the final output, or when
    only changing the output path for the final output.

    The ocr model (or None) is passed in as arguments, so it can be reused
    across multiple runs. This is done because in gui mode the model does not need to be loaded
    repeatedly for single image runs, making iterations much faster, due to the slow
    initializations of the model taking several seconds.

    When setting the output dir to None, intermediate steps are saved to disk, offering
    additional visualizations that aren't necessary when you aren't viewing image details.

    :param image_objects: The image objects to process. These contain the image paths.
    :param split_files: The split files information.
    :param target_outputs: The target outputs to reach, eg. the clean mask and image.
        Multiple outputs are only supported for those within the same step, as it's meant to accommodate
        the full run parameters.
    :param output_dir: The directory to save the outputs to. If None, the output remains in the cache directory.
    :param config: The config to use.
    :param ocr_processor: The ocr models to use.
    :param progress_callback: A callback to report progress to the gui.
    :param batch_metadata: The metadata for the batch, used for the post action arguments.
    :param abort_flag: A flag that is set to True when the thread should abort.
    :param debug: If True, debug messages are printed.
    """

    # We will check before each step and during each iteration of the step to see if the thread should abort.
    # The check is NOT done right before checking the outputs to avoid throwing away perfectly fine work.
    # This means that the abort check shouldn't interrupt the processing step on an image, leaving it in a
    # dirty state.
    def check_abortion() -> None:
        """
        Check if the thread should abort.
        If so, signal abortion to the progress callback and raise an AbortException.
        """
        if abort_flag.get():
            progress_callback.emit(
                ost.ProgressData(
                    0,
                    [],
                    ost.Step.output,
                    ost.ProgressType.aborted,
                )
            )
            raise wt.Abort()

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.text_detection,
            ost.ProgressType.start,
        )
    )

    # Create an independent copy of the profile, since this instance is shared and mutated across threads.
    # Due to the checksum only being generated at the end, we don't want to erroneously claim this output
    # matches a profile that was changed in the meantime.
    profile = deepcopy(config.current_profile)

    cache_dir = config.get_cleaner_cache_dir()

    # Pick out the highest output we need to reach.
    # Other outputs are only considered if they are in the same step, used for final output runs.
    target_output = max(target_outputs)

    # Get the text detector model path from the config.
    cuda = torch.cuda.is_available()
    text_detector_model_path = config.get_model_path(cuda)

    def step_needs_to_be_rerun_closure(current_step: ost.Step) -> Callable[[imf.ImageFile], bool]:
        """
        Check if any of the outputs for the given image object need to be rerun.

        :return: A function that takes an image object and returns True if the step needs to be rerun for it.
        """

        def step_changed(image_object: imf.ImageFile) -> bool:
            nonlocal profile
            nonlocal target_outputs
            nonlocal current_step
            # holy shit this is cursed I need to think about it more.
            # I think I got it sorted now...
            # Pretty sure it works :)
            # It didn't.
            # Now it is fr fr

            target_outputs_in_the_current_step: tuple[ost.Output] = tuple(
                filter(lambda o: o in ost.step_to_output[current_step], reversed(target_outputs))
            )

            if target_outputs_in_the_current_step:
                # We need to check each output we care about to know if we gotta rerun the step.
                return any(
                    image_object.outputs[o].is_changed(profile)
                    for o in target_outputs_in_the_current_step
                )
            else:
                # In this case we just need to know if the step's representative is complete.
                representative: ost.Output = ost.get_output_representing_step(current_step)
                return image_object.outputs[representative].is_changed(profile)

        return step_changed

    def update_output(image_object: imf.ImageFile, current_output: ost.Output) -> None:
        """
        Update the output of the given image object.
        Check if the file actually exists, and if it does, update the output path.

        :param image_object: The image object to update.
        :param current_output: The output to update.
        """
        nonlocal profile, cache_dir

        _path_gen = ost.OutputPathGenerator(image_object.path, cache_dir, image_object.uuid)
        path = _path_gen.for_output(current_output)

        if path.is_file():
            image_object.outputs[current_output].update(path, profile)

    # ============================================== Text Detection ==============================================

    check_abortion()

    step_text_detector_images = tuple(
        filter(step_needs_to_be_rerun_closure(ost.Step.text_detection), image_objects)
    )

    if step_text_detector_images:
        logger.info(
            f"Running text detection AI model for {len(step_text_detector_images)} images..."
        )

        visualize_raw_boxes = ost.Output.raw_boxes in target_outputs

        try:
            ctm.model2annotations_gui(
                profile.general,
                profile.text_detector,
                text_detector_model_path,
                step_text_detector_images,
                cache_dir,
                no_text_detection=target_output == ost.Output.input,
                visualize_raw_boxes=visualize_raw_boxes,
                partial_progress_data=partial(
                    ost.ProgressData,
                    len(step_text_detector_images),
                    target_outputs,
                    ost.Step.text_detection,
                ),
                progress_callback=progress_callback if target_output != ost.Output.input else None,
                abort_flag=abort_flag,
            )
        except wt.Abort:
            # Send the progress callback a signal that the process was aborted,
            # and then raise the exception again to stop the processing.
            check_abortion()

        for image_obj in step_text_detector_images:
            for output in ost.step_to_output[ost.Step.text_detection]:
                update_output(image_obj, output)

        progress_callback.emit(
            ost.ProgressData(
                len(step_text_detector_images),
                target_outputs,
                ost.Step.text_detection,
                ost.ProgressType.textDetection_done,
            )
        )

    # ============================================== Preprocessing ==============================================

    check_abortion()

    if target_output > ost.get_output_representing_step(ost.Step.text_detection):
        step_preprocessor_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.preprocessor), image_objects)
        )

        if step_preprocessor_images:
            logger.info(f"Running preprocessing for {len(step_preprocessor_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_preprocessor_images),
                    target_outputs,
                    ost.Step.preprocessor,
                    ost.ProgressType.begin_step,
                )
            )

            ocr_analytics: list[st.OCRAnalytic] = []
            # Find all the json files associated with the images.
            for image_obj in step_preprocessor_images:
                check_abortion()
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                ocr_analytic = pp.prep_json_file(
                    path_gen.raw_json,
                    preprocessor_conf=profile.preprocessor,
                    cache_masks=target_output in ost.step_to_output[ost.Step.preprocessor]
                    or output_dir is None,
                    ocr_engine_factory=ocr_processor if profile.preprocessor.ocr_enabled else None,
                )

                if ocr_analytic is not None:
                    ocr_analytics.append(ocr_analytic)

                progress_callback.emit(
                    ost.ProgressData(
                        len(step_preprocessor_images),
                        target_outputs,
                        ost.Step.preprocessor,
                        ost.ProgressType.incremental,
                    )
                )

            # Update the outputs of the image objects, skipping Output.ocr.
            for image_obj in step_preprocessor_images:
                update_output(image_obj, ost.Output.initial_boxes)
                update_output(image_obj, ost.Output.final_boxes)
                update_output(image_obj, ost.Output.clean_json)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_preprocessor_images),
                    target_outputs,
                    ost.Step.preprocessor,
                    ost.ProgressType.analyticsOCR,
                    (ocr_analytics, profile.preprocessor.ocr_max_size),
                )
            )

    # ============================================== Masker ==============================================

    check_abortion()

    if target_output > ost.get_output_representing_step(ost.Step.preprocessor):
        step_masker_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.masker), image_objects)
        )

        if step_masker_images:
            logger.info(f"Running masker for {len(step_masker_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_masker_images),
                    target_outputs,
                    ost.Step.masker,
                    ost.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = []
            for image_obj in step_masker_images:
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                json_files.append(path_gen.clean_json)

            # Pack all the arguments into a dataclass.
            outputs_that_need_masks = (
                ost.Output.box_mask,
                ost.Output.cut_mask,
                ost.Output.mask_layers,
                ost.Output.mask_overlay,
            )
            # If any of out targeted outputs need to have intermediate masks generated,
            # we need to show the masks. Or if we explicitly want to show the masks.
            need_to_show_masks = output_dir is None or any(
                output in outputs_that_need_masks for output in target_outputs
            )

            data = [
                st.MaskerData(
                    json_file,
                    cache_dir,
                    profile.masker,
                    extract_text=ost.Output.isolated_text in target_outputs,
                    show_masks=need_to_show_masks,
                    debug=debug,
                )
                for json_file in json_files
            ]

            masker_analytics_raw = []

            # Check the size of the required pool, run sequentially if the size is 1.
            core_limit = profile.masker.max_threads
            if core_limit == 0:
                core_limit = os.cpu_count()
            pool_size = min(core_limit, len(data))

            if pool_size > 1:
                with Pool(processes=pool_size) as pool:
                    for analytic in pool.imap(ma.mask_page, data):
                        check_abortion()
                        masker_analytics_raw.extend(analytic)

                        progress_callback.emit(
                            ost.ProgressData(
                                len(step_masker_images),
                                target_outputs,
                                ost.Step.masker,
                                ost.ProgressType.incremental,
                            )
                        )
            else:
                for data_obj in data:
                    check_abortion()
                    analytic = ma.mask_page(data_obj)
                    masker_analytics_raw.extend(analytic)

                    progress_callback.emit(
                        ost.ProgressData(
                            len(step_masker_images),
                            target_outputs,
                            ost.Step.masker,
                            ost.ProgressType.incremental,
                        )
                    )

            for image_obj in step_masker_images:
                for output in ost.step_to_output[ost.Step.masker]:
                    update_output(image_obj, output)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_masker_images),
                    target_outputs,
                    ost.Step.masker,
                    ost.ProgressType.analyticsMasker,
                    (masker_analytics_raw, copy(profile.masker)),  # Make a copy to avoid a mutex.
                )
            )

    # ============================================== Denoiser ==============================================

    check_abortion()

    if (
        target_output > ost.get_output_representing_step(ost.Step.masker)
        and profile.denoiser.denoising_enabled
    ):
        step_denoiser_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.denoiser), image_objects)
        )

        if step_denoiser_images:
            logger.info(f"Running denoiser for {len(step_denoiser_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_denoiser_images),
                    target_outputs,
                    ost.Step.denoiser,
                    ost.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = []
            for image_obj in step_denoiser_images:
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                json_files.append(path_gen.mask_data_json)

            # Pack all the arguments into a dataclass.
            data = [
                st.DenoiserData(
                    json_file,
                    cache_dir,
                    profile.denoiser,
                    debug,
                )
                for json_file in json_files
            ]

            denoise_analytics_raw = []

            # Check the size of the required pool, run sequentially if the size is 1.
            core_limit = profile.denoiser.max_threads
            if core_limit == 0:
                core_limit = os.cpu_count()
            pool_size = min(core_limit, len(data))

            if pool_size > 1:
                with Pool() as pool:
                    for analytic in pool.imap(dn.denoise_page, data):
                        check_abortion()
                        denoise_analytics_raw.append(analytic)

                        progress_callback.emit(
                            ost.ProgressData(
                                len(step_denoiser_images),
                                target_outputs,
                                ost.Step.denoiser,
                                ost.ProgressType.incremental,
                            )
                        )
            else:
                for data_obj in data:
                    check_abortion()
                    analytic = dn.denoise_page(data_obj)
                    denoise_analytics_raw.append(analytic)

                    progress_callback.emit(
                        ost.ProgressData(
                            len(step_denoiser_images),
                            target_outputs,
                            ost.Step.denoiser,
                            ost.ProgressType.incremental,
                        )
                    )

            for image_obj in step_denoiser_images:
                for output in ost.step_to_output[ost.Step.denoiser]:
                    update_output(image_obj, output)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_denoiser_images),
                    target_outputs,
                    ost.Step.denoiser,
                    ost.ProgressType.analyticsDenoiser,
                    (
                        denoise_analytics_raw,
                        profile.denoiser.noise_min_standard_deviation,
                        profile.masker.mask_max_standard_deviation,
                    ),
                )
            )

    # ============================================== Inpainting ==============================================

    check_abortion()

    if (
        target_output > ost.get_output_representing_step(ost.Step.denoiser)
        and profile.inpainter.inpainting_enabled
    ):
        step_inpainting_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.inpainter), image_objects)
        )

        if step_inpainting_images:
            logger.info(f"Running inpainting for {len(step_inpainting_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_inpainting_images),
                    target_outputs,
                    ost.Step.inpainter,
                    ost.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = []
            for image_obj in step_inpainting_images:
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                json_files.append((path_gen.clean_json, path_gen.mask_data_json))

            # Check the inpainting model is available.
            if not md.is_inpainting_downloaded(config):
                raise FileNotFoundError(tr("Inpainting model not found."))

            inpainter_model = ip.InpaintingModel(config)

            # Pack all the arguments into a dataclass.
            data = [
                st.InpainterData(
                    page_json_file,
                    mask_json_file,
                    cache_dir,
                    profile.masker,
                    profile.denoiser,
                    profile.inpainter,
                    debug,
                )
                for page_json_file, mask_json_file in json_files
            ]

            # Single threaded due to model loading overhead.
            inpaint_analytics_raw = []
            for data_obj in data:
                check_abortion()
                analytic = ip.inpaint_page(data_obj, inpainter_model)
                inpaint_analytics_raw.append(analytic)

                progress_callback.emit(
                    ost.ProgressData(
                        len(step_inpainting_images),
                        target_outputs,
                        ost.Step.inpainter,
                        ost.ProgressType.incremental,
                    )
                )

            for image_obj in step_inpainting_images:
                for output in ost.step_to_output[ost.Step.inpainter]:
                    update_output(image_obj, output)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_inpainting_images),
                    target_outputs,
                    ost.Step.inpainter,
                    ost.ProgressType.analyticsInpainter,
                    (
                        inpaint_analytics_raw,
                        profile.inpainter.min_inpainting_radius,
                        profile.inpainter.max_inpainting_radius,
                    ),
                )
            )

    # ============================================== Final Output ==============================================

    logger.info(f"Finished processing {len(image_objects)} images.")

    check_abortion()

    if target_output == ost.Output.write_output:

        handle_merging_splits(image_objects, split_files, profile, cache_dir)
        batch_metadata.set_input_paths_from_files(image_objects)

        progress_callback.emit(
            ost.ProgressData(
                len(image_objects),
                target_outputs,
                ost.Step.output,
                ost.ProgressType.begin_step,
            )
        )

        # Pack all the arguments in batches.
        data = [
            (
                image_obj.path,
                image_obj.export_path,
                cache_dir,
                image_obj.uuid,
                target_outputs,
                output_dir,
                profile.general.preferred_file_type,
                profile.general.preferred_mask_file_type,
                profile.denoiser.denoising_enabled,
                profile.general.layered_export,
            )
            for image_obj in image_objects
        ]

        # Check the size of the required pool, run sequentially if the size is 1.
        core_limit = profile.general.max_threads_export
        if core_limit == 0:
            core_limit = os.cpu_count()
        pool_size = min(core_limit, len(data))

        if pool_size > 1:
            with Pool(processes=pool_size) as pool:
                for output_files in pool.imap_unordered(ie.copy_to_output_batched, data):
                    check_abortion()
                    batch_metadata.output_files.extend(output_files)

                    progress_callback.emit(
                        ost.ProgressData(
                            len(image_objects),
                            target_outputs,
                            ost.Step.output,
                            ost.ProgressType.incremental,
                        )
                    )
        else:
            for args in data:
                check_abortion()
                output_files = ie.copy_to_output(*args)
                batch_metadata.output_files.extend(output_files)

                progress_callback.emit(
                    ost.ProgressData(
                        len(image_objects),
                        target_outputs,
                        ost.Step.output,
                        ost.ProgressType.incremental,
                    )
                )

        # For bulk exports, we need a common parent directory.
        if not output_dir.is_absolute():
            common_parent = hp.common_path_parent(
                [image_object.export_path for image_object in image_objects],
            )
            output_dir = common_parent / output_dir

        if profile.general.layered_export == LayeredExport.PSD_BULK:
            file_path = ie.bundle_psd(
                output_dir,
                cache_dir,
                [image_object.path for image_object in image_objects],
                [image_object.uuid for image_object in image_objects],
            )
            if file_path is not None:
                batch_metadata.output_files = [file_path]

        # Clean up merged files to prevent a storage leak
        if split_files:
            # These files start with merged_
            for file in cache_dir.glob("merger_*"):
                file.unlink()

    else:
        # We aren't writing output, but we may still need to update the metadata at least.
        # For that, we will need the input files, so we'd prefer to use the re-merged files.

        # If we need to merge files, we do it by replacing the split objects with the merged objects.
        # But we don't actually want to delete the image objects, so we create a shallow copy of the list.
        # Otherwise the review options, which share the image objects list, would have messed up
        # image objects as a side effect, then pass those on to the next run of the cleaner, resulting
        # in the poor thing being unable to handle the merged files.
        # We don't want the merged image objects to leak beyond the scope of this function.
        image_objects_export = image_objects.copy()
        handle_merging_splits(image_objects_export, split_files, profile, cache_dir)
        batch_metadata.set_input_paths_from_files(image_objects_export)

    batch_metadata.calculate_output_parents()

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.output,
            ost.ProgressType.end,
        )
    )


def perform_ocr(
    image_objects: list[imf.ImageFile],
    split_files: dict[Path, list[imf.ImageFile]],
    output_file: Path | None,
    csv_output: bool,
    config: cfg.Config,
    ocr_engine_factory: ocr.OCREngineFactory | None,
    progress_callback: ost.ProgressSignal,
    batch_metadata: gst.BatchMetadata,
    abort_flag: wt.SharableFlag,
    debug: bool = False,
):
    """
    Perform OCR on all the images in the given image objects list.
    If a path is given, the output is written to that path.
    Output is either a plain text file or csv.

    The ocr model (or None) is passed in as arguments, so it can be reused
    across multiple runs. This is done because in gui mode the model does not need to be loaded
    repeatedly for single image runs, making iterations much faster, due to the slow
    initializations of the model taking several seconds.

    When setting the output dir to None, intermediate steps are saved to disk, offering
    additional visualizations that aren't necessary when you aren't viewing image details.

    :param image_objects: The image objects to process. These contain the image paths.
    :param split_files: The split files information.
    :param output_file: The directory to save the outputs to. If None, nothing is written.
    :param csv_output: If True, the output is written as a csv file.
    :param config: The config to use.
    :param ocr_engine_factory: The ocr processors to use.
    :param progress_callback: A callback to report progress to the gui.
    :param batch_metadata: The metadata for the batch, used for the post action arguments.
    :param abort_flag: A flag that is set to True when the thread should abort.
    :param debug: If True, debug messages are printed.
    """

    def check_abortion() -> None:
        """
        Check if the thread should abort.
        If so, signal abortion to the progress callback and raise an AbortException.
        """
        if abort_flag.get():
            progress_callback.emit(
                ost.ProgressData(
                    0,
                    [],
                    ost.Step.output,
                    ost.ProgressType.aborted,
                )
            )
            raise wt.Abort()

    target_outputs = [ost.Output.ocr]

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.text_detection,
            ost.ProgressType.start,
        )
    )

    # Create an independent copy of the profile, since this instance is shared and mutated across threads.
    # Due to the checksum only being generated at the end, we don't want to erroneously claim this output
    # matches a profile that was changed in the meantime.
    # Also, we will be editing this profile to get OCR working.
    profile = deepcopy(config.current_profile)

    # Make sure OCR is enabled.
    profile.preprocessor.ocr_enabled = True
    # Make sure the max size is infinite, so no boxes are skipped in the OCR process.
    profile.preprocessor.ocr_max_size = 10**10
    # Set the OCR blacklist pattern to match everything, so all text gets reported in the analytics.
    profile.preprocessor.ocr_blacklist_pattern = ".*"

    cache_dir = config.get_cleaner_cache_dir()

    # Get the text detector model path from the config.
    cuda = torch.cuda.is_available()
    text_detector_model_path = config.get_model_path(cuda)

    def step_needs_to_be_rerun_closure(current_step: ost.Step) -> Callable[[imf.ImageFile], bool]:
        """
        Check if any of the outputs for the given image object need to be rerun.

        :return: A function that takes an image object and returns True if the step needs to be rerun for it.
        """

        def step_changed(image_object: imf.ImageFile) -> bool:
            nonlocal profile
            nonlocal target_outputs
            nonlocal current_step

            target_outputs_in_the_current_step: tuple[ost.Output] = tuple(
                filter(lambda o: o in ost.step_to_output[current_step], reversed(target_outputs))
            )

            if target_outputs_in_the_current_step:
                # We need to check each output we care about to know if we gotta rerun the step.
                return any(
                    image_object.outputs[o].is_changed(profile)
                    for o in target_outputs_in_the_current_step
                )
            else:
                # In this case we just need to know if the step's representative is complete.
                representative: ost.Output = ost.get_output_representing_step(current_step)
                return image_object.outputs[representative].is_changed(profile)

        return step_changed

    def update_output(image_object: imf.ImageFile, current_output: ost.Output) -> None:
        """
        Update the output of the given image object.
        Check if the file actually exists, and if it does, update the output path.

        :param image_object: The image object to update.
        :param current_output: The output to update.
        """
        nonlocal profile, cache_dir

        _path_gen = ost.OutputPathGenerator(image_object.path, cache_dir, image_object.uuid)
        path = _path_gen.for_output(current_output)

        if path.is_file():
            image_object.outputs[current_output].update(path, profile)

    # ============================================== Text Detection ==============================================

    check_abortion()

    step_text_detector_images = tuple(
        filter(step_needs_to_be_rerun_closure(ost.Step.text_detection), image_objects)
    )

    if step_text_detector_images:
        logger.info(
            f"Running text detection AI model for {len(step_text_detector_images)} images..."
        )
        try:
            ctm.model2annotations_gui(
                profile.general,
                profile.text_detector,
                text_detector_model_path,
                step_text_detector_images,
                cache_dir,
                no_text_detection=False,
                visualize_raw_boxes=True,
                partial_progress_data=partial(
                    ost.ProgressData,
                    len(step_text_detector_images),
                    target_outputs,
                    ost.Step.text_detection,
                ),
                progress_callback=progress_callback,
                abort_flag=abort_flag,
            )
        except wt.Abort:
            # Send the progress callback a signal that the process was aborted,
            # and then raise the exception again to stop the processing.
            check_abortion()

        # Update the outputs of the image objects.
        for image_obj in step_text_detector_images:
            for output in ost.step_to_output[ost.Step.text_detection]:
                update_output(image_obj, output)

    # ============================================== Preprocessing ==============================================

    check_abortion()

    logger.info(f"Running preprocessing for {len(image_objects)} images...")

    progress_callback.emit(
        ost.ProgressData(
            len(image_objects),
            target_outputs,
            ost.Step.preprocessor,
            ost.ProgressType.begin_step,
        )
    )

    ocr_analytics: list[st.OCRAnalytic] = []
    # Find all the json files associated with the images.
    for image_obj in image_objects:
        check_abortion()
        path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
        ocr_analytic = pp.prep_json_file(
            path_gen.raw_json,
            preprocessor_conf=profile.preprocessor,
            cache_masks=False,
            ocr_engine_factory=ocr_engine_factory,
            cache_masks_ocr=True,
            performing_ocr=True,
        )

        if ocr_analytic is not None:
            ocr_analytics.append(ocr_analytic)

        progress_callback.emit(
            ost.ProgressData(
                len(image_objects),
                target_outputs,
                ost.Step.preprocessor,
                ost.ProgressType.incremental,
            )
        )

    # Update only the raw boxes, the rest are tainted by the forced profile changes.
    for image_obj in image_objects:
        update_output(image_obj, ost.Output.initial_boxes)

    logger.info(f"Finished processing {len(image_objects)} images.")

    # ============================================== Final Output ==============================================

    check_abortion()

    # In the cleaning process we want to avoid messing with the image_objects list if we aren't exporting,
    # since we share it with the review options object, which will then reuse it later in turn for export.
    # With OCR however, we don't use the image objects list for export (just the analytics),
    # but we do use it when showing the image in the OCR editor/reviewer, in which case this
    # cross-contamination is explicitly necessary to have the reviewer show the merged image for
    # OCR editing, as is expected with the merged analytics containing the text.
    handle_merging_splits(image_objects, split_files, profile, cache_dir, for_ocr=True)
    handle_merging_ocr_splits(split_files, profile, ocr_analytics)

    check_abortion()

    # Output the OCRed text from the analytics.
    text_out = ocr.format_output(
        ocr_analytics,
        csv_output,
        (tr("filename"), tr("startx"), tr("starty"), tr("endx"), tr("endy"), tr("text")),
    )

    text_out = text_out.strip("\n \t")

    if output_file is not None:
        try:
            output_file.parent.mkdir(parents=True, exist_ok=True)
            output_file.write_text(text_out, encoding="utf-8")
            text_out += "\n\n" + tr("Saved detected text to {output_file}").format(
                output_file=output_file
            )
        except OSError:
            text_out += "\n\n" + tr("Failed to write detected text to {output_file}").format(
                output_file=output_file
            )
            gu.show_exception(None, tr("Save Failed"), tr("Failed to write detected text to file."))

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.preprocessor,
            ost.ProgressType.outputOCR,
            (text_out, ocr_analytics),
        )
    )

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.output,
            ost.ProgressType.end,
        )
    )


def handle_merging_splits(
    image_objects: list[imf.ImageFile | imf.MergedImageFile],
    split_files: dict[Path, list[imf.ImageFile]],
    profile: cfg.Profile,
    cache_dir: Path,
    for_ocr: bool = False,
):
    """
    If merging the splits, merge them into a new image object and remove the split objects.

    :param image_objects: The image objects to process.
    :param split_files: The split files information.
    :param profile: The profile to use.
    :param cache_dir: The cache directory to use.
    :param for_ocr: If True, the merge is for OCR, and we forge some output data as well.
    """
    if not split_files:
        return

    if profile.general.merge_after_split:
        logger.info("Merging split files...")

        for split_file in split_files:
            split_objects = split_files[split_file]

            if not split_objects:
                logger.error(f"Split file {split_file} has no split objects.")
                continue

            # Create a new image object with the split objects.
            merged_object = imf.MergedImageFile(split_objects, cache_dir, for_ocr)

            # Add the new object to the list of image objects.
            image_objects.append(merged_object)

            # Remove the split objects from the list of image objects.
            for split_object in split_objects:
                image_objects.remove(split_object)


def handle_merging_ocr_splits(
    split_files: dict[Path, list[imf.ImageFile]],
    profile: cfg.Profile,
    ocr_analytics: list[st.OCRAnalytic],
):
    """
    If merging the splits, merge them into a new image object and remove the split objects.

    :param split_files: The split files information.
    :param profile: The profile to use.
    :param ocr_analytics: The OCR analytics to merge.
    """
    if not split_files:
        return

    if profile.general.merge_after_split:
        logger.info("Merging split ocr...")

        for split_file in split_files:
            split_objects = split_files[split_file]

            if not split_objects:
                logger.error(f"Split file {split_file} has no split objects.")
                continue

            segment_paths = [split_object.path for split_object in split_objects]
            split_from = split_objects[0].split_from
            st.merge_ocr_analytics(split_from, segment_paths, ocr_analytics)
